"""
Library for simulating space-based VLBI missions (spacevlbi)

Copyright 2024 Ben Hudson

Licensed under the GNU General Public License, Version 3.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.gnu.org/licenses/gpl-3.0.en.html

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# TimeLoop.py
#
# The primary simulation function of spacevlbi, executing all major functionality
# of the tool.
#
# @author: BenHudson - 22/09/2024

from spacevlbi.Orbit import OrbitPropagation, SatGroundAccess
from spacevlbi.Attitude import AttitudePropagation
from spacevlbi.Constraints import ObsLimits
from spacevlbi.Observation import Baselines
import numpy as np
from astropy.timeseries import TimeSeries
from astropy import units as u
from astropy.coordinates import SkyCoord,GCRS
from astropy.coordinates import solar_system_ephemeris, get_body
solar_system_ephemeris.set("jpl")

###############################################################################
# TimeLoop
###############################################################################

def TimeLoop(initTime, simLength, timeStep, spaceTelescopes, groundTelescopes,\
             groundStations, frequency, sourceRa, sourceDec, dutyCycle, \
             intTime, allsky):
    """Iterate through time loop and perform space VLBI simulation.

    :param initTime: Simulation start datetime, defaults to None
    :type initTime: str
    :param simLength: Simulation duration in seconds, defaults to None
    :type simLength: int
    :param timeStep: Simulation time step in seconds, defaults to None
    :type timeStep: int
    :param spaceTelescopes: Array of SpaceTelescope objects, defaults to None
    :type spaceTelescopes: list
    :param groundTelescopes: Array of GroundTelescope objects, defaults to None
    :type groundTelescopes: list
    :param groundStations: Array of GroundStation objects, defaults to None
    :type groundStations: list
    :param frequency: Observation frequency in Hz, defaults to None
    :type frequency: float
    :param sourceRa: Right ascension of target source in degrees, defaults to None
    :type sourceRa: float
    :param sourceDec: Declination of target source in degrees, defaults to None
    :type sourceDec: float
    :param dutyCycle: Time from start of one integration to the next in seconds, 
        defaults to None
    :type dutyCycle: int
    :param intTime: Integration time of instrument, defaults to None
    :type intTime: int
    :param allsky: Calculate all-sky coverage? Defaults to None
    :type allsky: bool
    :return: Array of SpaceTelescope objects
    :rtype: list
    :return: Array of GroundTelescope objects
    :rtype: list
    :return: Array of GroundStation objects
    :rtype: list
    :return: Timeseries of simulation time, defaults to None
    :rtype: list
    """
    
    # Create time series for simulation
    simTime = TimeSeries(time_start=initTime, time_delta=timeStep * u.s,\
                         n_samples=np.floor(simLength/timeStep))
    
    # Initialise parameters for modelling integration time and duty cycle
    lastObservation = 0;
    obsDuration = 0
    
    # Execute time loop
    for i in range(1,int(np.floor(simLength/timeStep))):
        # Display simulation progress
        print(str(i+1) + " / " + str(int((np.floor(simLength/timeStep)))))
        
        # Calculate ground telescope positions in ECI
        if groundTelescopes:
            for j in range(len(groundTelescopes)):
                posECEF = groundTelescopes[j].ecefPosition 
                # Transform ECEF TO ECI
                ecef = SkyCoord(x=posECEF[0,0],y=posECEF[0,1],z=posECEF[0,2],\
                                unit='m',frame='itrs',representation_type=\
                                'cartesian', obstime = simTime.time[i])
                eci = (ecef.transform_to(GCRS)).cartesian
                # Update GroundTelescope objects
                groundTelescopes[j].eciPosition = np.vstack((groundTelescopes[j]. \
                                eciPosition, np.array([eci.x.value, eci.y.value, \
                                eci.z.value]) << u.m))
                
        # Calculate Sun and Moon positions using SPICE kernels
        sun = get_body('sun', simTime.time[i], ephemeris='de432s')
        sun = (sun.transform_to(GCRS)).cartesian
        rSun = np.array([sun.x.value, sun.y.value, sun.z.value])*1000 \
                            << u.m
        moon = get_body('moon', simTime.time[i], ephemeris='de432s')
        moon = (moon.transform_to(GCRS)).cartesian
        rMoon = np.array([moon.x.value, moon.y.value, moon.z.value])*1000 \
                            << u.m
                
        # Spacecraft Orbit Propagation
        spaceTelescopes = OrbitPropagation(spaceTelescopes, simTime.time[i],\
                                           rSun, rMoon, i, timeStep)
        
        # Spacecraft Attitude Propagation
        spaceTelescopes = AttitudePropagation(spaceTelescopes, rSun, \
                            rMoon, sourceRa, sourceDec, i, timeStep)
        
        # Calculate Range / Elevation from Spacecraft to Ground Stations
        spaceTelescopes, groundStations = SatGroundAccess(spaceTelescopes, \
                            groundStations, simTime.time[i])
        
        # Calculate limits on observations for each station
        spaceTelescopes, groundTelescopes = ObsLimits(spaceTelescopes, \
                            groundTelescopes, groundStations, sourceRa,\
                            sourceDec, rSun, rMoon)
            
        # Calculate VLBI baselines
        if (i*timeStep - lastObservation >= dutyCycle) and (obsDuration <= \
                            intTime):
            spaceTelescopes, groundTelescopes = Baselines(i, spaceTelescopes, \
            groundTelescopes, sourceRa, sourceDec, frequency,simLength, \
            timeStep, allsky)
            obsDuration = obsDuration + timeStep
            if obsDuration >= intTime:
                lastObservation = i * timeStep - intTime
                obsDuration = 0

    return spaceTelescopes, groundTelescopes, groundStations, simTime



"""
Library for simulating space-based VLBI missions (spacevlbi)

Copyright 2024 Ben Hudson

Licensed under the GNU General Public License, Version 3.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.gnu.org/licenses/gpl-3.0.en.html

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# ExampleSpaceTelescope.py
#
# An example instance of the SpaceTelescope class. The modelled spacecraft is
# representative of the preliminary concept of operations of the Black Hole 
# Explorer (BHEX). More detailed elements of the design defined here are for
# example only and are not intended to accurately represent BHEX's final design
# - https://www.blackholeexplorer.org/
#
# @author: BenHudson - 22/09/2024

from spacevlbi import Station
import numpy as np

def BaselineBHEX(initTime):
    """Example SpaceTelescope object implementation for the Black Hole Explorer
    (BHEX).

    :param initTime: Simulation start datetime, defaults to None
    :type initTime: str    
    :return: sc: SpaceTelescope object representing BHEX
    :rtype sc: SpaceTelescope
    """
    
###############################################################################
#   Environment properties
###############################################################################

    # Orbit perturbation models used in the propagation of the spacecraft's 
    # orbit. If set to 1, the perturbation will be modelled.
    gravityJ2 = 1  # Earth gravity field J2 harmonic
    gravityJ3 = 1  # Earth gravity field J3 harmonic
    atmosDrag = 1  # Atmospheric drag, exponential density model
    solarPress = 1  # Solar radiation pressure
    solarFlux = 1367  # W/m^2, solar flux from Sun
    gravityLuniSolar = 1  # Luni-Solar point model gravity perturbations
    
###############################################################################
#   General spacecraft properties
###############################################################################

    name = "BHEX";
    mass = 300;  # Spacecraft mass, kg
    areaDrag = 3.5;  # Spacecraft drag area, m^2
    areaSolar = 3.5;  # Spacecraft solar area, m^2
    cD = 2.2;  # Spacecraft drag coefficient
    cR = 2.2;  # Spacecraft solar reflectivity coefficient
    
###############################################################################
#   Orbit Configuration
###############################################################################

    sma = 26563.88  # semi-major axis, km
    ecc = 0.001  # Eccentricity
    inc = 90  # Inclination, deg
    ra = 247.7  # Right ascension of the ascending node, deg
    aop = 0  # Argument of perigee in degrees
    ta = 0  #  Starting true anomaly in degrees
    
###############################################################################
#   Attitude configuration
###############################################################################

    # Body-fixed axis to be pointed at target source. Should be the same as 
    # RadioPayload antennaBoresight to point antenna at target
    pointingVector = np.array([0,0,1]) 
    # Constraint vector in body frame. Must be perpendicular to pointingVector.
    # The direction in which this axis is pointed is defined by the RollAngle
    # parameter below
    constraintVector = np.array([0,1,0])
    # Roll angle of space telescope in degrees about the pointingVector. 
    # Please see diagram in the Guide section of the Documentation for clarity.
    # rollAngle format: [time of transition 1 in sec, roll angle 1 in degrees,
    # etc.]
    rollAngle = [0, 0]
    
###############################################################################
#   Payload configuration
###############################################################################

    payloadName = "VLBI"
    antennaDiameter = 3.5  # antenna diameter, metres
    antennaBoresight = np.array([0,0,1])  # antenna boresight in body frame
    apertureEff = 1  # Aperture efficiency
    antSunExcl = 90   # Antenna - Sun exclusion angle, deg
    antEarthExcl = 5  # Antenna - Earth limb exclusion angle, deg
    antMoonExcl = 0  # Antenna - Moon exclusion angle, deg
    sysTemp = 4.5  # System noise temperature, Kelvin
    
    # Initialise RadioPayload object
    payload1 = Station.RadioPayload(payloadName, antennaDiameter, apertureEff,\
                 sysTemp, antennaBoresight, antSunExcl, \
                 antEarthExcl, antMoonExcl)
    radioPayloads = [payload1]

###############################################################################
#   Equipment Model configuration
###############################################################################
    
    strModel = 0 # Model star trackers?
    # Number of unblinded star trackers required for observation
    reqStarTrackers = 2
    radModel = 0  # Model radiators?
    panelModel = 0  # Model solar panels?
    commsModel = 0  # Model comms systems?
    
###############################################################################
#   Star tracker configuration
###############################################################################

    # Note. During the simulation, if a star tracker is blinded at the current
    # time step, an observation cannot be performed.
    strName1 = "STR1"
    # Star tracker boresight in body frame
    strBoresight1 = np.array([-0.476, -0.655, -0.589]) 
    strSunExcl1 = 30   # Star tracker - Sun exclusion angle, deg
    strEarthExcl1 = 30  # Star tracker - Earth limb exclusion angle, deg
    strMoonExcl1 = 0  # Star tracker - Moon exclusion angle, deg
    
    # Initialise StarTracker object
    starTracker1 = Station.StarTracker(strName1, strBoresight1, strSunExcl1, \
                                      strEarthExcl1, strMoonExcl1)
    
    strName2 = "STR2"
    # Star tracker boresight in body frame
    strBoresight2 = np.array([0.707, 0, -0.707])  
    strSunExcl2 = 30   # Star tracker - Sun exclusion angle, deg
    strEarthExcl2 = 30  # Star tracker - Earth limb exclusion angle, deg
    strMoonExcl2 = 0  # Star tracker - Moon exclusion angle, deg
    
    # Initialise StarTracker object
    starTracker2 = Station.StarTracker(strName2, strBoresight2, strSunExcl2, \
                                       strEarthExcl2, strMoonExcl2)
    
    starTrackers = [starTracker1, starTracker2]
    
###############################################################################
#   Solar panel configuration
###############################################################################
    
    panelName1 = "Solar Panel 1"
    panelNorm1 = np.array([0,0,-1])  # Normal vector in body frame
    # Initialise SolarPanel object
    panel1 = Station.SolarPanel(panelName1, panelNorm1)
    
    solarPanels = [panel1]
    
###############################################################################
#   Radiator surface configuration
###############################################################################

    # Note. During the simulation, if a radiator is blinded at the current
    # time step, an observation cannot be performed.
    radName = "Rad"
    radNorm = np.array([-0.809, -0.588, 0]);  # Normal vector in body frame
    radSunExcl = 40  # Radiator - Sun exclusion angle, deg
    radEarthExcl = 40  # Radiator - Earth exclusion angle, deg
    radMoonExcl = 40  # Radiator - Moon exclusion angle, deg
    # Initialise Radiator object
    rad1 = Station.Radiator(radName, radNorm, radSunExcl, radEarthExcl, \
                 radMoonExcl)
        
    radiators = [rad1]
    
###############################################################################
#   Comms configuration
###############################################################################
    
    commsName = "Optical"
    commsNorm = np.array([1,0,0])  # Normal vector in body frame
    commsFov = 90  # Beamwidth/gimbal limit (half angle from normal vector), deg
    groundReqObs = 1  # Is a ground station required insight for observations?
    # Initialise CommsSystem object
    comms1 = Station.CommsSystem(commsName,  commsNorm, commsFov, groundReqObs)
    
    commsSystems = [comms1]
    
###############################################################################
#   Generate SpaceTelescope
###############################################################################

    sc = Station.SpaceTelescope(name, mass, areaDrag, areaSolar, cD, cR, \
                 initTime, sma, ecc, inc, ra, aop, ta,  pointingVector, \
                 constraintVector, rollAngle, radioPayloads, starTrackers, \
                 reqStarTrackers, radiators, commsSystems, solarPanels, \
                 strModel, radModel, commsModel, panelModel, gravityJ2, \
                 gravityJ3, atmosDrag, solarPress, solarFlux, gravityLuniSolar)
        
    return sc


"""
Library for simulating space-based VLBI missions (spacevlbi)

Copyright 2024 Ben Hudson

Licensed under the GNU General Public License, Version 3.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.gnu.org/licenses/gpl-3.0.en.html

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# Optimisation.py
#
# Contains the function required to optimise the position of a spacecraft
# component with a specific Sun, Earth and/or Moon relationship. Optimisation
# provides vector along which a given component can be pointed in the 
# spacecraft's body-fixed frame that minimises the amount of times during a 
# simulation that a defined functional constraint is violated.
#
# @author: BenHudson - 22/09/2024

import numpy as np
from numpy.linalg import norm
from numpy import arctan, dot, degrees, arccos
from astropy import constants as const

def Optimisation(spaceTelescopes, sunExcl, earthExcl, \
                 moonExcl, telescopeSelect=0, direction="greaterthan"):
    """Iterate through the entire attitude sphere and determine the positions
    at which the defined Sun, Earth and Moon exclusion angles are violated the
    least during the simulation. This function can be used to find the optimal 
    location of specific spacecraft components (E.g. star trackers, radiators,
    etc.) to minimise the impact of their functional constraint on
    observations. The 'optimal' nature of the component must be able to be
    defined in terms of a Sun, Earth and/or Moon exclusion angle.

    :param spaceTelescopes: Array of SpaceTelescope objects, defaults to None
    :type spaceTelescopes: list
    :param telescopeSelect: Index of spaceTelescope array to optimise, defaults to 0
    :type telescopeSelect: int
    :param sunExcl: Sun exclusion angle in degrees. The direction parameter
        controls whether the Sun angle must be less than or greater than this
        value in order to perform an observation, defaults to None
    :type sunExcl: float
    :param earthExcl: Earth exclusion angle in degrees. The direction parameter
        controls whether the Earth angle must be less than or greater than this
        value in order to perform an observation, defaults to None
    :type earthExcl: float
    :param moonExcl:  Moon exclusion angle in degrees. The direction parameter
        controls whether the Moon angle must be less than or greater than this
        value in order to perform an observation, defaults to None
    :type moonExcl: float
    :param direction:  "lessthan" or "greaterthan". Whether the angle between 
        the unit normal vector and the celestial body should be greater than 
        or less than the earthExcl, sunExcl or moonExcl in order for an 
        observation to take place. E.g. for a star tracker, direction is set 
        to "greaterthan" and the sunExcl parameter is set to the exclusion angle of 
        the star tracker unit, defaults to "greaterthan"
    :type direction: str
    :return: List of unit vectors in the spacecraft body-fixed frame 
        and the associated percentage of the simulation for which the Sun, Earth and/or
        Moon exclusion angles were violated. I.e. smaller number is more optimal.
    :rtype: numpy.ndarray
    """
    
    if spaceTelescopes:
    
        # Extract space telescope and simulation length
        telescope = spaceTelescopes[telescopeSelect]
        simLength = len(spaceTelescopes[telescopeSelect].eciPosition)
        
        # Generate search space of unit vectors
        u = np.linspace(0, 2 * np.pi, 40)
        v = np.linspace(0, np.pi, 40)
        x = np.outer(np.cos(u), np.sin(v))
        y = np.outer(np.sin(u), np.sin(v))
        z = np.outer(np.ones(np.size(u)), np.cos(v))
        [m,n] = x.shape
        
        # Initialise results matrix
        results = np.zeros((m*n, 4))
        resultsCount = 0
        
        # Iterate through attitude sphere positions
        for i in range(m):
            for j in range(n):
                # Generate unit vector
                unitVec = np.array([x[i,j], y[i,j], z[i,j]])
                # Initialise status array
                status = np.zeros((simLength,1))
                for k in range(1,simLength):
                    # Extract Sun, Moon and Earth positions in body-fixed frame
                    # from simulation run
                    sunBody = telescope.sunBody[k,:].value
                    moonBody = telescope.moonBody[k,:].value
                    earthBody = telescope.earthBody[k,:].value
                    
                    # Calculate limb angles
                    rSun = telescope.rSun[k,:].value
                    rMoon = telescope.rMoon[k,:].value
                    moonRad = 1737.4e3
                    position = telescope.eciPosition[k,:].value
                    earthLimb = arctan(const.R_earth.value / norm(position))
                    sunLimb = arctan(const.R_sun.value / norm(rSun))
                    moonLimb = arctan(moonRad / norm(rMoon))
                    
                    # Calculate angles between unit vector and Earth, Sun and Moon
                    # limbs
                    sunAngle = degrees(abs(arccos(dot(unitVec, sunBody)/ \
                        (norm(unitVec)*norm(sunBody))))-sunLimb)
                    moonAngle = degrees(abs(arccos(dot(unitVec, moonBody)/ \
                        (norm(unitVec)*norm(moonBody))))-moonLimb)
                    earthAngle = degrees(abs(arccos(dot(unitVec, earthBody)/ \
                        (norm(unitVec)*norm(earthBody))))-earthLimb)
                        
                    #print(sunAngle)
                    #print(moonAngle)
                    #print(earthAngle)
                        
                    if direction == "lessthan":
                        if (sunAngle < sunExcl) and (earthAngle < earthExcl) and \
                            (moonAngle < moonExcl):
                            status[k] = 0
                        else:
                            status[k] = 1
                    elif direction == "greaterthan":
                        if (sunAngle > sunExcl) and (earthAngle > earthExcl) and \
                            (moonAngle > moonExcl):
                            status[k] = 0
                        else:
                            status[k] = 1
                    else:
                        print("Invalid direction parameter")
                        
                # Calculate for how many timesteps Sun, Earth and Moon
                # exclusion angles were not violated for
                results[resultsCount, 0:3] = unitVec
                results[resultsCount, 3] = ((len(status)-1-\
                                    np.count_nonzero(status))/(len(status)-1))*100
                resultsCount = resultsCount + 1
        
        # Sort results placing optimal configurations at the top
        sorted_indices = np.argsort(-results[:,-1])
        results = results[sorted_indices]
        
        # Return sorted list of results and print optimal position with mimimal
        # impact on observations
        print("Optimal position vector: " + str(results[0,0:3]))
        print("Fitness (% of simulation constraint is not violated): " + \
              str(round(results[0,3],2)) + "%")
        print("See function output 'results' for a full list of optimal positions")
    
    else:
        print("Cannot perform optimisation, no space telescopes are modelled")
        results = 0
        
    return results
        
    
    
    
    




"""
Library for simulating space-based VLBI missions (spacevlbi)

Copyright 2024 Ben Hudson

Licensed under the GNU General Public License, Version 3.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.gnu.org/licenses/gpl-3.0.en.html

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# Orbit.py
#
# Contains the functions required to model a space telescope's orbit in spacevlbi.
#
# @author: BenHudson - 22/09/2024

import numpy as np
from numpy.linalg import norm
from numpy import degrees, arccos
from astropy.coordinates import SkyCoord,GCRS,ITRS
from astropy import units as u
from poliastro.bodies import Earth, Moon, Sun
from poliastro.twobody.propagation import CowellPropagator
from poliastro.core.propagation import func_twobody
from poliastro.core.perturbations import J2_perturbation, J3_perturbation, \
    atmospheric_drag_exponential
from poliastro.constants import rho0_earth, H0_earth

###############################################################################
# OrbitPropagation
###############################################################################

def OrbitPropagation(spaceTelescopes, time, rSun, rMoon, i, timeStep):
    """Propagate space telescopes' orbits using poliastro Orbit functionality.
    Propagation currently includes following perturbations: Earth J2, J3
    harmonics, luni-solar third-body, atmospheric drag and solar radiation pressure. 
    Propulsive burns in the form of a low-thrust, long-duration electric 
    propulsion system can also be modelled. Additional perturbing forces can be 
    added by editing the Force() function.

    :param spaceTelescopes: Array of SpaceTelescope objects, defaults to None
    :type spaceTelescopes: list 
    :param time: Current time in simulation, defaults to None
    :type time: str
    :param rSun: Sun position vector in ECI frame in metres, defaults to None
    :type rSun: list
    :param rMoon: Moon position vector in ECI frame in metres, defaults to None
    :type rMoon: list
    :param i: Current time step, defaults to None
    :type i: int
    :param timeStep: simulation time step in seconds, defaults to None
    :type timeStep: int
    :return: Array of spaceTelescope objects
    :rtype: list
    """

    # Iterate through spaceTelescopes
    if spaceTelescopes:
        for j in range(len(spaceTelescopes)):
            
            # Extract initial orbit from space telescope
            if i == 0:
                initPosition = spaceTelescopes[j].orbit
            else:
                initPosition = spaceTelescopes[j].orbitLast
                
            # Extract spacecraft properties
            mass = spaceTelescopes[j].mass
            areaDrag = spaceTelescopes[j].areaDrag * 1e-6  # Convert to km^2
            cD = spaceTelescopes[j].cD
            areaSolar = spaceTelescopes[j].areaSolar * 1e-6  # Convert to km^2
            cR = spaceTelescopes[j].cR
            gravityJ2 = spaceTelescopes[j].gravityJ2
            gravityJ3 = spaceTelescopes[j].gravityJ3
            atmosDrag = spaceTelescopes[j].atmosDrag
            solarPress = spaceTelescopes[j].solarPress
            solarFlux = spaceTelescopes[j].solarFlux
            gravityLuniSolar = spaceTelescopes[j].gravityLuniSolar
            
            duration = timeStep
            # Propagate orbit using Cowell's method and defined perturbations
            currentPosition = initPosition.propagate(duration << u.s, \
                                method=CowellPropagator(f=Force_Wrapper(rSun,\
                                rMoon,mass,areaDrag,cD,areaSolar,cR, gravityJ2, \
                                gravityJ3, atmosDrag, solarPress, \
                                solarFlux, gravityLuniSolar)))
                
            # Convert propagated orbit to ECI
            posNow = np.array(currentPosition.r*1000) << u.m
            velNow = np.array(currentPosition.v*1000) << (u.m / u.s)
            # Convert ECI to ECEF
            eci = SkyCoord(x=posNow[0],y=posNow[1],z=posNow[2],\
                           unit='m',frame='gcrs',representation_type='cartesian', \
                               obstime = time)
            ecef = (eci.transform_to(ITRS)).cartesian
            ecefNow = np.array([ecef.x.value, ecef.y.value, ecef.z.value]) << u.m
            
            # Update spaceTelescope object with new position
            spaceTelescopes[j].eciPosition = np.vstack((spaceTelescopes[j].eciPosition, posNow))
            spaceTelescopes[j].eciVelocity = np.vstack((spaceTelescopes[j].eciVelocity, velNow))
            spaceTelescopes[j].ecefPosition = np.vstack((spaceTelescopes[j].ecefPosition, ecefNow))
            
            spaceTelescopes[j].orbitLast = currentPosition
            
    return spaceTelescopes


###############################################################################
# Perturbations
###############################################################################

def Force_Wrapper(rSun, rMoon, mass, areaDrag, cD, areaSolar, cR, gravityJ2=1, \
                  gravityJ3=1, atmosDrag=1, solarPress=1, solarFlux=1367, \
                  gravityLuniSolar=1):
    def Force(t0, u_, k):
        """Force model for performing orbit propagation.
    
        :param t0: Current simulation time in seconds, defaults to None
        :type t0: float 
        :param u_: Current spacecraft state vector in ECI, defaults to None
        :type u_: numpy.ndarray
        :param k: Gravitational parameter of central body, defaults to None
        :type k: float
        :param rSun: Sun position vector in ECI frame in metres, defaults to None
        :type rSun: list
        :param rMoon: Moon position vector in ECI frame in metres, defaults to None
        :type rMoon: list
        :param mass: Spacecraft mass in kg, defaults to None
        :type mass: float
        :param areaDrag: Spacecraft drag area in km^2, defaults to None
        :type areaDrag: float
        :param cD: Spacecraft drag coefficient, defaults to None
        :type cD: float
        :param areaSolar: Spacecraft solar area in km^2, defaults to None
        :type areaSolar: float
        :param cR: Spacecraft reflectivity coefficient, defaults to None
        :type cR: float
        :param gravityJ2: Model Earth gravity harmonic J2? Defaults to 1
        :type gravityJ2: BOOL
        :param gravityJ3: Model Earth gravity harmonic J3? Defaults to 1
        :type gravityJ3: BOOL
        :param atmosDrag: Model atmospheric drag? Defaults to 1
        :type atmosDrag: BOOL
        :param solarPress: Model solar radiation pressure? Defaults to 1
        :type solarPress: BOOL
        :param solarFlux: Solar flux in W/m^2, defaults to 1367
        :type solarFlux: float
        :param gravityLuniSolar: Model luni-solar point model gravity? Defaults to 1
        :type gravityLuniSolar: BOOL
        :return: Acceleration vector imparted by perturbations in m/s^2
        :rtype: numpy.ndarray
        """
    
        # Two body acceleration
        du_kep = func_twobody(t0, u_, k)
        accel = du_kep
        
        # Sun - Moon parameters
        kMoon = Moon.k.to(u.km**3 / u.s**2).value
        kSun = Sun.k.to(u.km**3 / u.s**2).value
        
        rSat = u_[0:3]
        rEarthMoon = rMoon.value / 1000
        rSatMoon = rEarthMoon - rSat
        
        rEarthSun = rSun.value / 1000
        rSatSun = rEarthSun - rSat
        
        if gravityJ2 == 1:
            # J2 acceleration
            ax, ay, az = J2_perturbation(t0, u_, k, J2=Earth.J2.value, \
                            R=Earth.R.to(u.km).value)
            accel = accel + np.array([0, 0, 0, ax, ay, az])
            
        
        if gravityJ3 == 1:
            # J3 acceleration
            ax1, ay1, az1 = J3_perturbation(t0, u_, k, J3=Earth.J3.value, \
                            R=Earth.R.to(u.km).value)
            accel = accel + np.array([0, 0, 0, ax1, ay1, az1])            
            
        if gravityLuniSolar == 1:
            # Luni-Solar acceleration
            ax2, ay2, az2 = kMoon*((rSatMoon/(rSatMoon**3))-(rEarthMoon/(rEarthMoon**3)))
            ax3, ay3, az3 = kSun*((rSatSun/(rSatSun**3))-(rEarthSun/(rEarthSun**3)))
            
            accel = accel + np.array([0, 0, 0, ax2, ay2, az2]) + \
                np.array([0, 0, 0, ax3, ay3, az3])  
            
        if atmosDrag == 1:
            # Drag acceleration
            R = Earth.R.to(u.km).value
            rho0 = rho0_earth.to(u.kg / u.km**3).value  # kg/km^3
            H0 = H0_earth.to(u.km).value
            A_m = areaDrag / mass
            ax4, ay4, az4 = atmospheric_drag_exponential(t0, u_, k, R=R, C_D=cD, \
                                                          A_over_m=A_m, H0=H0, rho0=rho0)
            accel = accel + np.array([0, 0, 0, ax4, ay4, az4])
            
        if solarPress == 1:
            # Solar pressure acceleration
            flux = solarFlux * 1e-6  # W/km^2
            p = flux / 299792.458
            ax5, ay5, az5 = -p*cR*(areaSolar/mass)*(rSatSun/np.linalg.norm(rSatSun))
            accel = accel + np.array([0, 0, 0, ax5, ay5, az5])
        
        return accel
    
    return Force


###############################################################################
# SatGroundAccess
###############################################################################

def SatGroundAccess(spaceTelescopes, groundStations, time):
    """Calculate range and elevation from ground stations to space telescopes,
    measured in topocentric frame.

    :param spaceTelescopes: Array of SpaceTelescope objects, defaults to None
    :type spaceTelescopes: list
    :param groundStations: Array of GroundStation objects, defaults to None
    :type groundStations: list
    :param time: Current simulation time, defaults to None
    :type time: str
    :return: Array of SpaceTelescope objects
    :rtype: list
    :return: Array of GroundStation objects
    :rtype: list
    """
           
    # Iterate through ground stations, transform coordinates to ECI and 
    # calculate range and elevation
    if groundStations:
        for k in range(len(groundStations)):
            # Initialise range and elevation vectors
            satRange = np.zeros((1,len(spaceTelescopes))) << u.m
            elevation = np.zeros((1,len(spaceTelescopes)))
            
            # Iterate through space telescopes
            if spaceTelescopes:
                for j in range(len(spaceTelescopes)):
                    # Extract spacecraft position
                    spacecraftPos = spaceTelescopes[j].eciPosition[-1,:].reshape((3,1))
                    # Extract ground station position
                    groundECEF = groundStations[k].ecefPosition.reshape((3,1))
                    # Convert ground station position to ECI
                    ecef =SkyCoord(x=groundECEF[0],y=groundECEF[1],z=groundECEF[2],\
                                    unit='m',frame='itrs',representation_type=\
                                    'cartesian', obstime = time)
                    eci = (ecef.transform_to(GCRS)).cartesian    
                    eciPosition = np.array([eci.x.value, eci.y.value, \
                                    eci.z.value]).reshape((3,1)) << u.m
                    # Update ground station position
                    groundStations[k].eciPosition = np.vstack((groundStations[k].eciPosition, \
                                    eciPosition.reshape((1,3))))
                    # Calculate satellite - ground station range magnitude
                    satGroundRange = norm(spacecraftPos - eciPosition)
                    # Calculate satellite - ground station range vector
                    satGroundVector = (spacecraftPos - eciPosition).reshape((3,1))
        
                    # Calculate elevation of spacecraft w.r.t ground station
                    elRad = np.dot(satGroundVector[:,0], eciPosition[:,0]) / \
                                    (norm(satGroundVector)*norm(eciPosition))
                    el = degrees((np.pi/2) - arccos(elRad.value))
                    
                    # Assign to temporary vector
                    satRange[0,j] = satGroundRange
                    elevation[0,j] = el
                
                # Pad satRange and satElev parameters of ground stations to match the 
                # number of spacecraft.
                if len(groundStations[k].satRange) == 1:
                    groundStations[k].satRange = np.append(groundStations[k].satRange,\
                                    np.zeros((1,len(spaceTelescopes)-1)))
                if len(groundStations[k].satElev) == 1:
                    groundStations[k].satElev = np.append(groundStations[k].satElev, \
                                    np.zeros((1,len(spaceTelescopes)-1)))
                    
                # Update ground stations
                groundStations[k].satRange = np.vstack((groundStations[k].satRange,\
                                                        satRange))
                groundStations[k].satElev = np.vstack((groundStations[k].satElev,\
                                                        elevation))
                    
                #print(elevation)
            
    return spaceTelescopes, groundStations




"""
Library for simulating space-based VLBI missions (spacevlbi)

Copyright 2024 Ben Hudson

Licensed under the GNU General Public License, Version 3.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.gnu.org/licenses/gpl-3.0.en.html

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# Constraints.py
#
# Implementation of the functional constraints that apply to the space telescopes
# and the ground telescopes that will impact when observations can be performed
# in spacevlbi.
#
# @author: BenHudson - 22/09/2024

from astropy import constants as const
from astropy import units as u
import numpy as np
from numpy.linalg import norm
from numpy import vstack, dot, degrees, radians, cos, sin, arccos, arctan

###############################################################################
# ObsLimits
###############################################################################

def ObsLimits(spaceTelescopes, groundTelescopes, groundStations, sourceRa, \
              sourceDec, rSun, rMoon):
    """Calculate whether any of the functional constraints are stopping an
    observation from being performed at the current time step.

    :param spaceTelescopes: Array of SpaceTelescope objects, defaults to None
    :type spaceTelescopes: list
    :param groundTelescopes: Array of GroundTelescope objects, defaults to None
    :type groundTelescopes: list
    :param groundStations: Array of GroundStation objects, defaults to None
    :type groundStations: list
    :param sourceRa: Right ascension of target source in degrees, defaults to None
    :type sourceRa: float
    :param sourceDec: Declination of target source in degrees, defaults to None
    :type sourceDec: float
    :param rSun: Sun position vector in ECI frame in metres, defaults to None
    :type rSun: numpy.ndarray
    :param rMoon: Moon position vector in ECI frame in metres, defaults to None
    :type rMoon: numpy.ndarray
    :return: Array of SpaceTelescope objects
    :rtype: list
    :return: Array of GroundTelescope objects
    :rtype: list
    """
    
    # Iterate through space telescopes and calculate functional constraints
    if spaceTelescopes:
        for j in range(len(spaceTelescopes)):
            position = spaceTelescopes[j].eciPosition[-1,:] << u.m
            # Is the target source in view of the spacecraft?
            vis = SourceVisibility(sourceRa, sourceDec, position)
            spaceTelescopes[j].sourceVisibility = vstack((spaceTelescopes[j].sourceVisibility, \
                                vis))
                
            # Calculate angle of Earth limb from spacecraft position vector
            earthLimb = arctan(const.R_earth / norm(position))
            spaceTelescopes[j].earthLimbAngle = vstack((spaceTelescopes[j].earthLimbAngle, \
                                earthLimb))
                
            # Calculate angle of Sun limb from spacecraft position vector
            satSun = rSun - position
            sunLimb = arctan(const.R_sun / norm(satSun))
            spaceTelescopes[j].sunLimbAngle = vstack((spaceTelescopes[j].sunLimbAngle, \
                                sunLimb))
                
            # Calculate angle of Moon limb from spacecraft position vector
            moonRad = 1737.4e3 << u.m
            satMoon = rMoon - position
            moonLimb = arctan(moonRad / norm(satMoon))
            spaceTelescopes[j].moonLimbAngle = vstack((spaceTelescopes[j].moonLimbAngle, \
                                moonLimb))
                
            # Update Sun and Moon vectors in SpaceTelescope
            spaceTelescopes[j].rSun = vstack((spaceTelescopes[j].rSun, satSun))
            spaceTelescopes[j].rMoon = vstack((spaceTelescopes[j].rMoon, satMoon))
            
            sunInertial = spaceTelescopes[j].sunInertial[-1,:]
            moonInertial = spaceTelescopes[j].moonInertial[-1,:]
            earthInertial = (-spaceTelescopes[j].eciPosition[-1,:]/ \
                                norm(spaceTelescopes[j].eciPosition[-1,:]))
                
            # Is the antenna pointing within Sun, Moon or Earth exclusion zones?
            radioPayloads = spaceTelescopes[j].radioPayloads
            if radioPayloads != []:
                for k in range(len(radioPayloads)):
                    antennaInertial = radioPayloads[k].antennaInertial[-1,:]
                
                    sunAngle = abs(arccos(dot(antennaInertial, sunInertial)/ \
                        (norm(antennaInertial)*norm(sunInertial)))) - sunLimb
                    moonAngle = abs(arccos(dot(antennaInertial, moonInertial)/ \
                        (norm(antennaInertial)*norm(moonInertial)))) - moonLimb
                    earthAngle = abs(arccos(dot(antennaInertial, earthInertial)/ \
                        (norm(antennaInertial)*norm(earthInertial))))-earthLimb
            
                    if degrees(sunAngle.value) < radioPayloads[k].antSunExcl:
                        radioPayloads[k].sunFlag = vstack((radioPayloads[k].sunFlag,0))
                    else:
                        radioPayloads[k].sunFlag = vstack((radioPayloads[k].sunFlag,1))
                    if degrees(moonAngle.value) < radioPayloads[k].antMoonExcl:
                        radioPayloads[k].moonFlag = vstack((radioPayloads[k].moonFlag,0))
                    else:
                        radioPayloads[k].moonFlag= vstack((radioPayloads[k].moonFlag,1))
                    if degrees(earthAngle.value) < radioPayloads[k].antEarthExcl:
                        radioPayloads[k].earthFlag = vstack((radioPayloads[k].earthFlag,0))
                    else:
                        radioPayloads[k].earthFlag = vstack((radioPayloads[k].earthFlag,1))
            else:
                radioPayloads[k].sunFlag = vstack((radioPayloads[k].sunFlag,1))
                radioPayloads[k].moonFlag = vstack((radioPayloads[k].moonFlag,1))
                radioPayloads[k].earthFlag = vstack((radioPayloads[k].earthFlag,1))
            spaceTelescopes[j].radioPayloads = radioPayloads
    
            # Are the star trackers pointing within Sun, Moon or Earth exclusion
            # angles?
            starTrackers = spaceTelescopes[j].starTrackers
            if starTrackers != [] and spaceTelescopes[j].strModel == 1:
                for k in range(len(starTrackers)):
                    strInertial = starTrackers[k].strInertial[-1,:]
                    strSunAngle = abs(arccos(dot(strInertial, sunInertial)/ \
                        (norm(strInertial)*norm(sunInertial))))
                    strMoonAngle = abs(arccos(dot(strInertial, moonInertial)/ \
                        (norm(strInertial)*norm(moonInertial))))
                    strEarthAngle = abs(arccos(dot(strInertial, earthInertial)/ \
                        (norm(strInertial)*norm(earthInertial)))) - earthLimb;
                    if (degrees(strSunAngle).value<starTrackers[k].strSunExcl) or \
                            (degrees(strMoonAngle).value<starTrackers[k].strMoonExcl) or \
                            (degrees(strEarthAngle.value)<starTrackers[k].strEarthExcl):
                       starTrackers[k].strBlindFlag = vstack((starTrackers[k].strBlindFlag,0))
                    else:
                       starTrackers[k].strBlindFlag = vstack((starTrackers[k].strBlindFlag,1))
                starTrackers[k].starTrackers = starTrackers
                    
            # Are the radiators pointing within Sun, Moon or Earth exclusion angles?
            radiators = spaceTelescopes[j].radiators
            if radiators != [] and spaceTelescopes[j].radModel == 1:
                for k in range(len(radiators)):   
                    radInertial = radiators[k].radInertial[-1,:]
                    radSunAngle = abs(arccos(dot(radInertial, sunInertial)/ \
                        (norm(radInertial)*norm(sunInertial))));
                    radMoonAngle = abs(arccos(dot(radInertial, moonInertial)/ \
                        (norm(radInertial)*norm(moonInertial))));
                    radEarthAngle = abs(arccos(dot(radInertial, earthInertial)/ \
                        (norm(radInertial)*norm(earthInertial)))) - earthLimb;
                    if (degrees(radSunAngle.value)<radiators[k].radSunExcl) or \
                            (degrees(radMoonAngle.value)<radiators[k].radMoonExcl) or \
                            (degrees(radEarthAngle.value)<radiators[k].radEarthExcl):
                       radiators[k].radBlindFlag = vstack((radiators[k].radBlindFlag,0))
                    else:
                       radiators[k].radBlindFlag = vstack((radiators[k].radBlindFlag,1))
                spaceTelescopes[j].radiators = radiators
    
            # Are there any ground station limitations on observations?
            commsSystems = spaceTelescopes[j].commsSystems
            if commsSystems != [] and spaceTelescopes[j].commsModel == 1:
                for k in range(len(commsSystems)):
                    # Is live downlink of the data being performed?
                    if commsSystems[k].groundReqObs == 1:
                        # Initialise angle parameters
                        angle = np.zeros((1, len(groundStations)))
                        insight = np.zeros((1, len(groundStations)))
                        # Iterate through ground stations
                        for l in range(len(groundStations)):
                            # Calculate elevation of ground station to spacecraft
                            groundECI = groundStations[l].eciPosition[-1,:]
                            el = groundStations[l].satElev[-1,j]
                            minEl = groundStations[l].minEl
                            
                            # Caculate angle between spacecraft comms system normal
                            # vector and the ground station
                            commsInertial = commsSystems[k].commsInertial[-1,:]
                            spaceGround = (groundECI - position)
                            spaceGround = spaceGround / norm(spaceGround)
                            commsGroundAngle = degrees(abs(arccos(dot(commsInertial, \
                                        spaceGround)/ (norm(commsInertial)* \
                                        norm(spaceGround))))).value
                            commsFov = commsSystems[k].commsFov
                            angle[0,l] = commsGroundAngle
        
                            # Sspacecraft must be above minEl and the comms terminal 
                            # beamwidth / gimbal limit must be pointed within commFOV
                            # of the ground station for a link to be maintained
                            if (el < 90 and el > minEl) and commsGroundAngle <= commsFov:
                                insight[0,l] = 1;
                            else:
                                insight[0,l] = 0;
                                
                        if len(commsSystems[k].commsGroundAngle) == 1:
                            commsSystems[k].commsGroundAngle = \
                                np.append(commsSystems[k].commsGroundAngle, \
                                        np.zeros((1,len(groundStations)-1)))
                        if len(commsSystems[k].groundStationInsight) == 1:
                            commsSystems[k].groundStationInsight = \
                                np.append(commsSystems[k].groundStationInsight, \
                                        np.zeros((1,len(groundStations)-1)))
                        
                        commsSystems[k].commsGroundAngle = \
                            vstack((commsSystems[k].commsGroundAngle, angle))
                        commsSystems[k].groundStationInsight = \
                            vstack((commsSystems[k].groundStationInsight, insight))
                spaceTelescopes[j].commsSystems = commsSystems
    
    if groundTelescopes:
        for j in range(len(groundTelescopes)):
            # Is the elevation between the ground station and target source
            # within limits?
            position = groundTelescopes[j].eciPosition[-1,:]
            el = Elevation(position, sourceRa, sourceDec);
            groundTelescopes[j].elevation = vstack((groundTelescopes[j].elevation,el));
            minEl = groundTelescopes[j].minEl
            if el < 90 and el > minEl:
                groundTelescopes[j].elevationFlag = vstack((groundTelescopes[j].\
                                                            elevationFlag,1));
            else:
                groundTelescopes[j].elevationFlag = vstack((groundTelescopes[j].\
                                                            elevationFlag,0));

    return spaceTelescopes, groundTelescopes


###############################################################################
# SourceVisibility
###############################################################################

def SourceVisibility(ra, dec, position):
    """Calculate whether source at given right ascension and declination is in
    view of a space telescope (i.e. not blocked by the Earth).

    :param ra: Right ascension of target source in degrees, defaults to None
    :type ra: float
    :param dec: Declination of target source in degrees, defaults to None
    :type dec: float
    :param position: Spacecraft ECI position vector in metres, defaults to None
    :type position: numpy.ndarray
    :return: Flag indicating whether source is in view of spacecraft.
        In View == 1, Obstructed == 0.
    :rtype: bool
    """
    
    R_Earth = const.R_earth
    # Calculate Earth-source unit vector
    sourceECI = np.array([sin(np.pi/2 - radians(dec)) * cos(radians(ra)), \
                          sin(np.pi/2 - radians(dec)) * sin(radians(ra)), \
                          cos(np.pi/2 - radians(dec))])
    earthSource = (sourceECI / np.linalg.norm(sourceECI)) << u.m
    
    if (np.dot(position.value, earthSource.value) < \
        -np.sqrt(np.linalg.norm(position.value)**2 - R_Earth.value**2)):
        visibility = 0;
    else:
        visibility = 1;
    
    return visibility

###############################################################################
# Elevation
###############################################################################

def Elevation(position, ra, dec):
    """Calculate elevation of a target source from the horizon at a ground 
    telescope location in the topocentric frame.

    :param position: GroundTelescope ECI position vector in metres, defaults to 
        None
    :param ra: Right ascension of target source in degrees, defaults to None
    :type ra: float
    :param dec: Declination of target source in degrees, defaults to None
    :type dec: float
    :type position: numpy.ndarray
    :return: Elevation of target source from horizon in degrees
    :rtype: float
    """
    
    # Calculate source vector in ECI
    sourceECI = np.array([cos(radians(dec)) * cos(radians(ra)), \
                          cos(radians(dec)) * sin(radians(ra)), \
                          sin(radians(dec))])
    earthSourceECI = (sourceECI / norm(sourceECI)) * const.kpc

    # Generate vector from ground telescope to source
    groundSource = earthSourceECI - position
    
    # Calculate elevation of source from ground telescope
    elRad = dot(groundSource, position)/(norm(groundSource)* norm(position));
    el = degrees(0.5*np.pi - arccos(elRad).value);
    
    return el

###############################################################################
# ObsMask
###############################################################################

def ObsMask(telescope1, telescope2):
    """For two specific antenna, calculate whether an observation can be
    performed at the current time step.

    :param telescope1: SpaceTelescope or GroundTelescope object, defaults to None
    :type telescope1: SpaceTelescope or GroundTelescope
    :param telescope2: SpaceTelescope or GroundTelescope object, defaults to None
    :type telescope2: SpaceTelescope or GroundTelescope
    :return: Flag indicating whether observation can be performed by 
        these two antenna at the current time step
    :rtype: bool
    """
    
    # Check telescope1 functional constraint flags
    if telescope1.__class__.__name__ == 'SpaceTelescope':
        
        radioPayloads = telescope1.radioPayloads
        starTrackers = telescope1.starTrackers
        radiators = telescope1.radiators
        commsSystems = telescope1.commsSystems
        
        radioFlag = 1
        if radioPayloads != []:    
            for j in range(len(radioPayloads)):
                sunFlag = radioPayloads[j].sunFlag[-1,:]
                earthFlag = radioPayloads[j].earthFlag[-1,:]
                moonFlag = radioPayloads[j].moonFlag[-1,:]
                radioFlag = radioFlag * sunFlag * earthFlag * moonFlag

        strFlag = 1
        strReq = telescope1.reqStarTrackers
        strCount = 0
        if starTrackers != [] and telescope1.strModel == 1:    
            for j in range(len(starTrackers)):
                blindFlag = starTrackers[j].strBlindFlag[-1,:]
                if blindFlag == 1:
                    strCount = strCount + 1
            
            if strCount < strReq:
                strFlag = 0
            else:
                strFlag = 1
        
        radFlag = 1
        if radiators != [] and telescope1.radModel == 1:    
            for j in range(len(radiators)):
                blindFlag = radiators[j].radBlindFlag[-1,:]
                radFlag = radFlag * blindFlag

        # Ground station visibility limits
        commsFlag = 1
        if commsSystems != [] and telescope1.commsModel == 1:    
            for j in range(len(commsSystems)):
                # Is live downlink of the data being performed?
                if commsSystems[j].groundReqObs == 1:
                    flag = any(commsSystems[j].groundStationInsight[-1,:])
                else:
                    flag = 1
                commsFlag = commsFlag * flag

        flag1 = radioFlag * strFlag * radFlag * commsFlag
    else:
        flag1 = 1;
    
    # Check telescope1 functional constraint flags
    if telescope2.__class__.__name__ == 'SpaceTelescope':
        
        radioPayloads = telescope2.radioPayloads
        starTrackers = telescope2.starTrackers
        radiators = telescope2.radiators
        commsSystems = telescope2.commsSystems
        
        radioFlag = 1
        if radioPayloads != []:    
            for j in range(len(radioPayloads)):
                sunFlag = radioPayloads[j].sunFlag[-1,:]
                earthFlag = radioPayloads[j].earthFlag[-1,:]
                moonFlag = radioPayloads[j].moonFlag[-1,:]
                radioFlag = radioFlag * sunFlag * earthFlag * moonFlag
                
        strFlag = 1
        strReq = telescope2.reqStarTrackers
        strCount = 0
        if starTrackers != [] and telescope2.strModel == 1:    
            for j in range(len(starTrackers)):
                blindFlag = starTrackers[j].strBlindFlag[-1,:]
                if blindFlag == 1:
                    strCount = strCount + 1
            
            if strCount < strReq:
                strFlag = 0
            else:
                strFlag = 1
        
        radFlag = 1
        if radiators != [] and telescope2.radModel == 1:    
            for j in range(len(radiators)):
                blindFlag = radiators[j].radBlindFlag[-1,:]
                radFlag = radFlag * blindFlag

        # Ground station visibility limits
        commsFlag = 1
        if commsSystems != [] and telescope2.commsModel == 1:    
            for j in range(len(commsSystems)):
                # Is live downlink of the data being performed?
                if commsSystems[j].groundReqObs == 1:
                    flag = any(commsSystems[j].groundStationInsight[-1,:])
                else:
                    flag = 1
                commsFlag = commsFlag * flag

        flag2 = radioFlag * strFlag * radFlag * commsFlag
    else:
        flag2 = 1;

    # If any flags are 0 then observation cannot take place
    obsFlag = flag1 * flag2

    return obsFlag
    



"""
Library for simulating space-based VLBI missions (spacevlbi)

Copyright 2024 Ben Hudson

Licensed under the GNU General Public License, Version 3.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.gnu.org/licenses/gpl-3.0.en.html

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# Observation.py
#
# Functions required to simulate VLBI observations of target source and calculate
# baselines in spacevlbi.
#
# @author: BenHudson - 22/09/2024

from spacevlbi.Constraints import ObsMask
from astropy import constants as const
import numpy as np
from numpy.linalg import norm
from numpy import dot, cross, radians, cos, sin
from spacevlbi import Constraints

###############################################################################
# Baselines
###############################################################################

def Baselines(i, spaceTelescopes, groundTelescopes, sourceRa, sourceDec, \
              frequency, simLength, timeStep, allsky):
    """Calculate baselines formed by an array of a given source (or evaluated
    for a range of locations across the celestial sphere if allsky == 1).
    Baselines that couldn't be formed due to the impact of a functional
    constraint are also calculated. 

    :param i: Current time step, defaults to None
    :type i: int
    :param spaceTelescopes: Array of SpaceTelescope objects, defaults to None
    :type spaceTelescopes: list
    :param groundTelescopes: Array of GroundTelescope objects, defaults to None
    :type groundTelescopes: list
    :param sourceRa: Right ascension of target source in degrees, defaults to None
    :type sourceRa: float
    :param sourceDec: Declination of target source in degrees, defaults to None
    :type sourceDec: float
    :param frequency: Observation frequency in Hz, defaults to None
    :type frequency: float
    :param simLength: Simulation duration in seconds, defaults to None
    :type simLength: int
    :param timeStep: Simulation time step in seconds, defaults to None
    :type timeStep: int
    :param allsky: Calculate all-sky coverage? Defaults to None
    :type allsky: bool    
    :return: Array of SpaceTelescope objects
    :rtype: list
    :return: Array of GroundTelescope objects
    :rtype: list
    """
    
    # Speed of light
    c = const.c

    # Calculate wavelength
    wavelength = c / frequency
    
    # Form raan and dec arrays if allsky is 1
    if allsky == 1:
        # Define celestial sphere ranges
        raan = range(0,360,30)
        declination = range(-90,120,30)
    else:
        raan = [sourceRa]
        declination = [sourceDec]
        
    # Iterate through right ascension range
    for r in range(len(raan)):
        # Iterate through declination range
        for d in range(len(declination)):
            
            ra = radians(raan[r])
            dec = radians(declination[d])
            
            # Convert source ra and dec to vector in ECI frame
            sourceECI = np.array([cos(dec) * cos(ra), \
                                 cos(dec) * sin(ra), \
                                 sin(dec)])
            sourceVec = (sourceECI/norm(sourceECI))
            # Calculate proj U and V directions
            projU = cross(np.array([0,0,1]),sourceVec)
            projU = projU/norm(projU)
            projV = -cross(projU, sourceVec)

            # Calculate baselines
            # Iterate through spaceTelescopes
            for j in range(len(spaceTelescopes)):
                
                # If spaceTelescope's baseline property has length 0, calculate
                # the required length and initialise zero array
                if len(spaceTelescopes[j].baselines) == 0:
                    rows = len(raan) * len(declination)
                    stationNo = len(spaceTelescopes) + len(groundTelescopes)
                    matrix = []
                    matrixLost = []
                    for _ in range(rows):
                        matrix.append(np.zeros((int(round(simLength/timeStep)),\
                                                int(stationNo)*3)))
                        matrixLost.append(np.zeros((int(round(simLength/timeStep)),\
                                                    int(stationNo)*3)))
                    spaceTelescopes[j].baselines  = matrix
                    spaceTelescopes[j].lostBaselines = matrixLost
                
                baselineCount = 1
                # Extract position vector of telescope
                eciVector1 = spaceTelescopes[j].eciPosition[-1,:]
                
                # Check whether source is insight of space telescope
                sourceVis1 = Constraints.SourceVisibility(raan[r], declination[d], \
                                              eciVector1)
            
                # Iterate through spaceTelescope array. Calculate baselines formed
                # with primary spaceTelescope of this iteration
                for l in range(len(spaceTelescopes)):
                    # Extract position vector of telescope
                    eciVector2 = spaceTelescopes[l].eciPosition[-1,:]
            
                    # Determine whether observation is possible with these two antenna
                    # based on operational limitations
                    if allsky == 0:
                        obsFlag = ObsMask(spaceTelescopes[j], spaceTelescopes[l])
                    else:
                        obsFlag = 1

                    # Check whether source is insight of space telescope
                    sourceVis2 = Constraints.SourceVisibility(raan[r], declination[d], \
                                                  eciVector2)
                    
                    # Form (u,v) points
                    u = dot((eciVector1-eciVector2)/wavelength, projU)/1e9
                    v = dot((eciVector1-eciVector2)/wavelength, projV)/1e9
                    vector = np.array([u.value,v.value,0]).reshape((3,1))
                    
                    if j != l:
                        # If any conditions prohibiting observations are true, zero
                        # baseline, else calculate baseline
                        if obsFlag == 0 and sourceVis1 == 1 and sourceVis2 == 1:
                            spaceTelescopes[j].baselines[r*len(declination)+d]\
                                [i,baselineCount*3-3:baselineCount*3] = [0,0,0]
                            # Add lost baselines to space telescope property
                            spaceTelescopes[j].lostBaselines[r*len(declination)+d]\
                                [i,baselineCount*3-3:baselineCount*3] = vector.reshape((1,3))
                        elif obsFlag == 1 and sourceVis1 == 1 and sourceVis2 == 1:
                            # Calculate uvw coordinates
                            spaceTelescopes[j].baselines[r*len(declination)+d]\
                                [i,baselineCount*3-3:baselineCount*3] = vector.reshape((1,3))
                            spaceTelescopes[j].lostBaselines[r*len(declination)+d]\
                                [i,baselineCount*3-3:baselineCount*3] = [0,0,0]
                        else:
                            spaceTelescopes[j].baselines[r*len(declination)+d]\
                                [i,baselineCount*3-3:baselineCount*3] = [0,0,0]
                            spaceTelescopes[j].lostBaselines[r*len(declination)+d]\
                                [i,baselineCount*3-3:baselineCount*3] = [0,0,0]
                    else:
                        spaceTelescopes[j].baselines[r*len(declination)+d]\
                            [i,baselineCount*3-3:baselineCount*3] = [0,0,0]
                        spaceTelescopes[j].lostBaselines[r*len(declination)+d]\
                            [i,baselineCount*3-3:baselineCount*3] = [0,0,0]
                    
                    # Increase baseline index
                    baselineCount = baselineCount + 1

                # Iterate through ground array
                for l in range(len(groundTelescopes)):
                    # Extract position vector of telescope
                    eciVector2 = groundTelescopes[l].eciPosition[-1,:]
            
                    # Determine whether observation is possible with these two antenna
                    # based on operational limitations
                    if allsky == 0:
                        obsFlag = ObsMask(spaceTelescopes[j], groundTelescopes[l])
                    else:
                        obsFlag = 1
                    
                    # Check whether source is insight of ground telescope
                    el = Constraints.Elevation(eciVector2, raan[r], \
                                            declination[d])
                    if el < 90 and el > groundTelescopes[l].minEl:
                        sourceVis2 = 1
                    else:
                        sourceVis2 = 0
                    
                    # Form (u,v) points
                    u = dot((eciVector1-eciVector2)/wavelength, projU)/1e9
                    v = dot((eciVector1-eciVector2)/wavelength, projV)/1e9
                    vector = np.array([u.value,v.value,0]).reshape((3,1))
                    
                    # If any conditions prohibiting observations are true, zero
                    # baseline, else calculate baseline
                    if obsFlag == 0 and sourceVis1 == 1 and sourceVis2 == 1:
                        spaceTelescopes[j].baselines[r*len(declination)+d]\
                            [i,baselineCount*3-3:baselineCount*3] = [0,0,0]
                        # Add lost baselines to space telescope property
                        spaceTelescopes[j].lostBaselines[r*len(declination)+d]\
                            [i,baselineCount*3-3:baselineCount*3] = vector.reshape((1,3))
                    elif obsFlag == 1 and sourceVis1 == 1 and sourceVis2 == 1:
                        # Calculate uvw coordinates
                        spaceTelescopes[j].baselines[r*len(declination)+d]\
                            [i,baselineCount*3-3:baselineCount*3] = vector.reshape((1,3))
                        spaceTelescopes[j].lostBaselines[r*len(declination)+d]\
                            [i,baselineCount*3-3:baselineCount*3] = [0,0,0];
                    else:
                        spaceTelescopes[j].baselines[r*len(declination)+d]\
                            [i,baselineCount*3-3:baselineCount*3] = [0,0,0]
                        spaceTelescopes[j].lostBaselines[r*len(declination)+d]\
                            [i,baselineCount*3-3:baselineCount*3] = [0,0,0]
                    
                    # Increase baseline index
                    baselineCount = baselineCount + 1

            # Iterate through groundTelescope array
            for j in range(len(groundTelescopes)):
                
                # If groundTelescope's baseline property has length 0, calculate
                # the required length and initialise zero array
                if len(groundTelescopes[j].baselines) == 0:
                    rows = len(raan) * len(declination)
                    stationNo = len(spaceTelescopes) + len(groundTelescopes)
                    matrix = []
                    matrixLost = []
                    for _ in range(rows):
                        matrix.append(np.zeros((int(round(simLength/timeStep)),int(stationNo)*3)))
                        matrixLost.append(np.zeros((int(round(simLength/timeStep)),int(stationNo)*3)))
                    groundTelescopes[j].baselines  = matrix
                    groundTelescopes[j].lostBaselines = matrixLost
                                                                 
                baselineCount = 1
                # Extract position vector of telescope
                eciVector1 = groundTelescopes[j].eciPosition[-1,:];
                
                # Check whether source is insight of ground telescope
                el = Constraints.Elevation(eciVector1, raan[r], \
                                        declination[d])
                if el < 90 and el > groundTelescopes[j].minEl:
                    sourceVis1 = 1
                else:
                    sourceVis1 = 0
            
                # Iterate through groundTelescopes array
                for l in range(len(groundTelescopes)):
                    # Extract position vector of telescope
                    eciVector2 = groundTelescopes[l].eciPosition[-1,:]
            
                    # Determine whether observation is possible with these two antenna
                    # based on operational limitations
                    if allsky == 0:
                        obsFlag = ObsMask(groundTelescopes[j], groundTelescopes[l])
                    else:
                        obsFlag = 1
                    
                    # Check whether source is insight of ground telescope
                    el = Constraints.Elevation(eciVector2, raan[r], \
                                            declination[d])
                    if el < 90 and el > groundTelescopes[l].minEl:
                        sourceVis2 = 1
                    else:
                        sourceVis2 = 0
                        
                    # Form (u,v) points
                    u = dot((eciVector1-eciVector2)/wavelength, projU)/1e9
                    v = dot((eciVector1-eciVector2)/wavelength, projV)/1e9
                    vector = np.array([u.value,v.value,0]).reshape((3,1))
                        
                    if j != l:
                        # If any conditions prohibiting observations are true, zero
                        # baseline, else calculate baseline
                        if obsFlag == 0 and sourceVis1 == 1 and sourceVis2 == 1:
                            groundTelescopes[j].baselines[r*len(declination)+d]\
                                [i,baselineCount*3-3:baselineCount*3] = [0,0,0]
                            # Add lost baselines to space telescope property
                            groundTelescopes[j].lostBaselines[r*len(declination)+d]\
                                [i,baselineCount*3-3:baselineCount*3] = vector.reshape((1,3))
                        elif obsFlag == 1 and sourceVis1 == 1 and sourceVis2 == 1:
                            # Calculate uvw coordinates
                            groundTelescopes[j].baselines[r*len(declination)+d]\
                                [i,baselineCount*3-3:baselineCount*3] = vector.reshape((1,3))
                            groundTelescopes[j].lostBaselines[r*len(declination)+d]\
                                [i,baselineCount*3-3:baselineCount*3] = [0,0,0]
                        else:
                            groundTelescopes[j].baselines[r*len(declination)+d]\
                            [i,baselineCount*3-3:baselineCount*3] = [0,0,0]
                            groundTelescopes[j].lostBaselines[r*len(declination)+d]\
                            [i,baselineCount*3-3:baselineCount*3] = [0,0,0]
                    else:
                        groundTelescopes[j].baselines[r*len(declination)+d]\
                        [i,baselineCount*3-3:baselineCount*3] = [0,0,0]
                        groundTelescopes[j].lostBaselines[r*len(declination)+d]\
                        [i,baselineCount*3-3:baselineCount*3] = [0,0,0]

                    # Increase baseline index
                    baselineCount = baselineCount + 1

                # Iterate through spaceStation array
                for l in range(len(spaceTelescopes)):
                    # Extract position vector of telescope
                    eciVector2 = spaceTelescopes[l].eciPosition[-1,:]
                    
                    # Determine whether observation is possible with these two antenna
                    # based on operational limitations
                    if allsky == 0:
                        obsFlag = ObsMask(groundTelescopes[j], spaceTelescopes[l])
                    else:
                        obsFlag = 1
                    
                    # Check whether source is insight of space telescope
                    sourceVis2 = Constraints.SourceVisibility(raan[r], declination[d], \
                                                  eciVector2)

                    # Form (u,v) points
                    u = dot((eciVector1-eciVector2)/wavelength, projU)/1e9
                    v = dot((eciVector1-eciVector2)/wavelength, projV)/1e9
                    vector = np.array([u.value,v.value,0]).reshape((3,1))
            
                    # If any conditions prohibiting observations are true, zero
                    # baseline, else calculate baseline
                    if obsFlag == 0 and sourceVis1 == 1 and sourceVis2 == 1:
                        groundTelescopes[j].baselines[r*len(declination)+d]\
                            [i,baselineCount*3-3:baselineCount*3] = [0,0,0]
                        # Add lost baselines to space telescope property
                        groundTelescopes[j].lostBaselines[r*len(declination)+d]\
                            [i,baselineCount*3-3:baselineCount*3] = vector.reshape((1,3))
                    elif obsFlag == 1 and sourceVis1 == 1 and sourceVis2 == 1:
                        # Calculate uvw coordinates
                        groundTelescopes[j].baselines[r*len(declination)+d]\
                            [i,baselineCount*3-3:baselineCount*3] = vector.reshape((1,3))
                        groundTelescopes[j].lostBaselines[r*len(declination)+d]\
                            [i,baselineCount*3-3:baselineCount*3] = [0,0,0]
                    else:
                        groundTelescopes[j].baselines[r*len(declination)+d]\
                            [i,baselineCount*3-3:baselineCount*3] = [0,0,0]
                        groundTelescopes[j].lostBaselines[r*len(declination)+d]\
                            [i,baselineCount*3-3:baselineCount*3] = [0,0,0]
                    
                    baselineCount = baselineCount + 1
           
    return spaceTelescopes, groundTelescopes


"""
Library for simulating space-based VLBI missions (spacevlbi)

Copyright 2024 Ben Hudson

Licensed under the GNU General Public License, Version 3.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.gnu.org/licenses/gpl-3.0.en.html

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# Attitude.py
#
# Contains the function required to propagate a space telescope's attitude state
# in spacevlbi.
#
# @author: BenHudson - 22/09/2024

import numpy as np
from numpy.linalg import norm, inv
from numpy import cross, matmul, vstack, dot, degrees, radians, cos, sin, arccos
from astropy import constants as const
from astropy import units as u

def AttitudePropagation(spaceTelescopes, rSun, rMoon, sourceRa, sourceDec, i, 
                        timeStep):
    """Propagate space telescope's attitude state. This function calculates the
    attitude matrix required to point the antenna at the target source. The
    constraint axis is pointed in a direction perpendicular to the antenna 
    direction, defined by the rollAngle parameter of a SpaceTelescope.

    :param spaceTelescopes: Array of SpaceTelescope objects, defaults to None
    :type spaceTelescopes: list
    :param rSun: Sun position vector in ECI frame in metres, defaults to None
    :type rSun: numpy.ndarray
    :param rMoon: Moon position vector in ECI frame in metres, defaults to None
    :type rMoon: numpy.ndarray
    :param sourceRa: Right ascension of target source in degrees, defaults to None
    :type sourceRa: float
    :param sourceDec: Declination of target source in degrees, defaults to None
    :type sourceDec: float
    :param i: Current time step, defaults to None
    :type i: int
    :param timeStep: simulation time step in seconds, defaults to None
    :type timeStep: int
    :return: Array of SpaceTelescope objects
    :rtype: list
    """
    
    # Iterate through space telescopes
    if spaceTelescopes:
        for j in range(len(spaceTelescopes)):
            # Calculate unit vector of source in spacecraft inertial frame
            sourceECI = np.array([cos(radians(sourceDec))*cos(radians(sourceRa)), \
                                  cos(radians(sourceDec))*sin(radians(sourceRa)), \
                                  sin(radians(sourceDec))])
            earthSourceECI = (sourceECI / norm(sourceECI)) * const.kpc
            # Extract spacecraft position vector
            rECI = spaceTelescopes[j].eciPosition[-1,:] << u.m
            rSource = (earthSourceECI - rECI)/np.linalg.norm(earthSourceECI - rECI)
            # Update space telescope object
            spaceTelescopes[j].sourceInertial = vstack((spaceTelescopes[j].sourceInertial, \
                                rSource))
                
            # Calculate Sun and Moon vectors with respect to spacecraft
            rSatSun = (rSun - rECI)/norm(rSun - rECI)
            rSatMoon = (rMoon - rECI)/norm(rMoon - rECI)
            spaceTelescopes[j].sunInertial = vstack((spaceTelescopes[j].sunInertial, \
                                rSatSun))
            spaceTelescopes[j].moonInertial = vstack((spaceTelescopes[j].moonInertial, \
                                rSatMoon))
                
            # Calculate attitude matrix to point antenna at target
            b1 = spaceTelescopes[j].pointingVector[0]  # Body pointing axis
            b2 = spaceTelescopes[j].constraintVector[0]  # Body constraint axis
            r1 = rSource  # Inertial pointing axis
            # Inertial constraint axis excluding roll angle
            r2Temp = cross(r1, b1) / norm(cross(r1, b1))
            # Line temporary constraint axis parallel with celestial north
            r2Temp = cross(-r2Temp, b1) / norm(cross(-r2Temp, b1))
            
            # Include roll angle. Find latest attitude transition
            rollAngle = spaceTelescopes[j].rollAngle
            lastAngleFound = 0
            theta = 0
            if len(rollAngle) == 2:
                theta = radians(spaceTelescopes[j].rollAngle[1])
            else:
                for k in range(1,int(len(rollAngle)/2)+1):
                    if ((i*timeStep) > rollAngle[k*-2]) and (lastAngleFound == 0):
                        theta = radians(spaceTelescopes[j].rollAngle[k*-2 + 1])
                        lastAngleFound = 1
            
            r2 = cos(theta)*r2Temp + sin(theta)*cross(r1, r2Temp) + (1 - cos(theta))* \
                 dot(r1, r2Temp) * r1
            # Calculate attitude matrix
            attMat = TRIAD(r1,r2,b1,b2)
            
            # Propagate spacecraft attitude, rotate spacecraft component vectors
            # Rotate antenna
            radioPayloads = spaceTelescopes[j].radioPayloads
            if radioPayloads != []:
                for k in range(len(radioPayloads)):
                    antennaInertial = matmul(inv(attMat), radioPayloads[k].antBoresight.reshape((3,1)))
                    radioPayloads[k].antennaInertial = vstack((radioPayloads[k].antennaInertial, \
                                    antennaInertial.reshape((1,3))))
                spaceTelescopes[j].radioPayloads = radioPayloads
            
            # Rotate star trackers
            starTrackers = spaceTelescopes[j].starTrackers
            if starTrackers != [] and spaceTelescopes[j].strModel == 1:
                for k in range(len(starTrackers)):
                    strInertial = matmul(inv(attMat), starTrackers[k].strBoresight.reshape((3,1)))
                    starTrackers[k].strInertial = vstack((starTrackers[k].strInertial, \
                                    strInertial.reshape((1,3))))
                spaceTelescopes[j].starTrackers = starTrackers
                    
            # Rotate solar panels and calculate beta angle
            solarPanels = spaceTelescopes[j].solarPanels
            if solarPanels != [] and spaceTelescopes[j].panelModel == 1:
                for k in range(len(solarPanels)):            
                    panelInertial = matmul(inv(attMat), solarPanels[k].panelNorm.reshape((3,1)))
                    solarPanels[k].solarPanelInertial = vstack((solarPanels[k].solarPanelInertial, \
                                    panelInertial.reshape((1,3))))
                    betaAngle = abs(arccos(dot(rSatSun, panelInertial) / (norm(rSatSun) \
                                    * norm(panelInertial))))
                    solarPanels[k].betaAngle = vstack((solarPanels[k].betaAngle,\
                                    degrees(betaAngle)))
                spaceTelescopes[j].solarPanels = solarPanels
               
            # Rotate radiator surface
            radiators = spaceTelescopes[j].radiators
            if radiators != [] and spaceTelescopes[j].radModel == 1:
                for k in range(len(radiators)):      
                    radInertial = matmul(inv(attMat), radiators[k].radNorm.reshape((3,1)))
                    radiators[k].radInertial = vstack((radiators[k].radInertial, \
                                    radInertial.reshape((1,3))))
                spaceTelescopes[j].radiators = radiators
                
            # Rotate comms terminal
            commsSystems = spaceTelescopes[j].commsSystems
            if commsSystems != [] and spaceTelescopes[j].commsModel == 1:
                for k in range(len(commsSystems)):
                    commsInertial = matmul(inv(attMat), commsSystems[k].commsNorm.reshape((3,1)))
                    commsSystems[k].commsInertial = vstack((commsSystems[k].commsInertial, \
                                    commsInertial.reshape((1,3))))
                spaceTelescopes[j].commsSystems = commsSystems
    
            # Rotate inertial vectors into spacecraft body frame
            sunBody = matmul(attMat, rSatSun.reshape((3,1)))
            spaceTelescopes[j].sunBody = vstack((spaceTelescopes[j].sunBody, \
                                sunBody.reshape((1,3))))
            moonBody = matmul(attMat, rSatMoon.reshape((3,1))) 
            spaceTelescopes[j].moonBody = vstack((spaceTelescopes[j].moonBody, \
                                moonBody.reshape((1,3))))
            earthBody = matmul(attMat, (-rECI / norm(rECI)).reshape((3,1)))
            spaceTelescopes[j].earthBody = vstack((spaceTelescopes[j].earthBody, \
                                earthBody.reshape((1,3))))
        
    return spaceTelescopes
    

def TRIAD(r1,r2,b1,b2):
    """TRIAD algorithm to calculate the attitude matrix from two pairs of body
    and inertial constraint vectors.

    :param r1: Inertial pointing vector, defaults to None
    :type r1: numpy.ndarray
    :param r2: Inertial constraint vector, defaults to None
    :type r2: numpy.ndarray
    :param b1: Body pointing vector, defaults to None
    :type b1: numpy.ndarray
    :param b2: Body constraint vector, defaults to None
    :type b2: numpy.ndarray
    :return: Attitude Matrix
    :rtype: numpy.ndarray
    """
    
    rCross = cross(r1,r2) / norm(cross(r1,r2))
    bCross = cross(b1,b2) / norm(cross(b1,b2))

    a1 = matmul(b1.reshape((3,1)), r1.reshape((1,3)))
    a2 = cross(b1, bCross).reshape((3,1))
    a3 = cross(r1, rCross).reshape((1,3))
    a4 = matmul(bCross.reshape((3,1)), rCross.reshape((1,3)))
        
    attMat = a1 + matmul(a2, a3) + a4
        
    return attMat




"""
Library for simulating space-based VLBI missions (spacevlbi)

Copyright 2024 Ben Hudson

Licensed under the GNU General Public License, Version 3.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.gnu.org/licenses/gpl-3.0.en.html

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# Figures.py
#
# Functions to generate figures to show results of a spacevlbi simulation.
#
# @author: BenHudson - 22/09/2024

from matplotlib import pyplot as plt
import numpy as np
from numpy import radians
from astropy import constants as const
import os
import geopandas
from shapely import Polygon

###############################################################################
# OrbitPlot
###############################################################################

def OrbitPlot(spaceTelescopes):
    """Plot orbits of space telescopes.

    :param spaceTelescopes: Array of SpaceTelescope objects, defaults to None
    :type spaceTelescopes: list
    """
    
    if spaceTelescopes:
    
        # Generate figure axis
        fig = plt.figure(figsize=(8,8))
        print("Generating orbit plot...")
    
        ax = fig.add_subplot(221, projection = '3d')
        earthColour = 'g'
        
        # Generate Earth sphere
        r = const.R_earth
        u, v = np.mgrid[0:2 * np.pi:30j, 0:np.pi:20j]
        x = r * np.cos(u) * np.sin(v)
        y = r * np.sin(u) * np.sin(v)
        z = r * np.cos(v)
        ax.plot_surface(x, y, z, cmap=plt.cm.YlGnBu_r)
        
        # Initialise maximum array
        maxPos = 0
        
        # Iterate through space telescopes and plot orbits
        for i in range(len(spaceTelescopes)):
            position = spaceTelescopes[i].eciPosition
            if np.max(position.value) > maxPos:
                maxPos = np.max(position.value)
            name = spaceTelescopes[i].name
            ax.plot(position[:,0], position[:,1], position[:,2], label=name)
        
        # Configure axes
        axisLimit = 1.1 * maxPos
        ax.set(xlim=(-axisLimit, axisLimit), ylim=(-axisLimit, axisLimit), \
               zlim=(-axisLimit, axisLimit))
        ax.set_xlabel('')
        ax.set_ylabel('')
        ax.set_zlabel('')
        ax.set_xticklabels([])
        ax.set_yticklabels([])
        ax.set_zticklabels([])
        ax.legend(loc="upper left")
        ax.set_aspect('equal', 'box')  
        
        ax = fig.add_subplot(222)
        # Iterate through space telescopes and plot orbits
        for i in range(len(spaceTelescopes)):
            position = spaceTelescopes[i].eciPosition
            ax.plot(position[:,0], position[:,1])
            
        # Add Earth circumference
        circle1 = plt.Circle((0, 0), const.R_earth.value, color= earthColour, \
                             fill=False)
        ax.add_artist(circle1)
        
        # Configure axes
        axisLimit = 1.1 * maxPos
        ax.set(xlim=(-axisLimit, axisLimit), ylim=(-axisLimit, axisLimit))
        ax.set_xlabel('x [km]')
        ax.set_ylabel('y [km]')
        ax.set_aspect('equal', 'box')  
        
        ax = fig.add_subplot(223)
        # Iterate through space telescopes and plot orbits
        for i in range(len(spaceTelescopes)):
            position = spaceTelescopes[i].eciPosition
            ax.plot(position[:,0], position[:,2])
            
        # Add Earth circumference
        circle1 = plt.Circle((0, 0), const.R_earth.value, color= earthColour,\
                             fill=False)
        ax.add_artist(circle1)
        
        # Configure axes
        axisLimit = 1.1 * maxPos
        ax.set(xlim=(-axisLimit, axisLimit), ylim=(-axisLimit, axisLimit))
        ax.set_xlabel('x [km]')
        ax.set_ylabel('z [km]')
        ax.set_aspect('equal', 'box')  
        
        ax = fig.add_subplot(224)
        # Iterate through space telescopes and plot orbits
        for i in range(len(spaceTelescopes)):
            position = spaceTelescopes[i].eciPosition
            name = spaceTelescopes[i].name
            ax.plot(position[:,1], position[:,2])
            
        # Add Earth circumference
        circle1 = plt.Circle((0, 0), const.R_earth.value, color= earthColour,\
                             fill=False)
        ax.add_artist(circle1)
        
        # Configure axes
        axisLimit = 1.1 * maxPos
        ax.set(xlim=(-axisLimit, axisLimit), ylim=(-axisLimit, axisLimit))
        ax.set_xlabel('y [km]')
        ax.set_ylabel('z [km]')
        ax.set_aspect('equal', 'box')  
        
        plt.show()    
        
        #python program to check if a directory exists
        path = "Outputs"
        # Check whether the specified path exists or not
        isExist = os.path.exists(path)
        if not isExist:
           # Create a new directory because it does not exist
           os.makedirs(path)
           print("Creating Outputs folder")
           
        fig.savefig('Outputs/Orbit.pdf', format='pdf', \
                    bbox_inches='tight')
    else:
        print("Cannot generate Orbit plot, no space telescopes are modelled")

###############################################################################
# UvPlot
###############################################################################

def UvPlot(spaceTelescopes, groundTelescopes, allsky, plotLost=1):
    """Plot (u,v) coverage of a target source(s) generated by the simulated 
    array.

    :param spaceTelescopes: Array of SpaceTelescope objects, defaults to None
    :type spaceTelescopes: list
    :param groundTelescopes: Array of GroundTelescope objects, defaults to None
    :type groundTelescopes: list
    :param plotLost: Plot baselines lost due to functional constraints? 
        Defaults to 1
    :type plotLost: bool
    :param allysky: Plot all-sky coverage? Should be equal to allsky parameter
        defined in the Timeloop() function
    :type allysky: bool
    """
    
    if allsky == 0:
        # Marker size
        m = 10
        
        # Calculate number of stations
        stationNo = len(spaceTelescopes) + len(groundTelescopes)
        if stationNo > 1:
    
            fig = plt.figure(figsize=(8,8))
            ax = fig.add_subplot()
            print("Generating (u,v) coverage plot...")
            
            # Plot u = v = 0 lines
            ax.plot([-200, 200], [0,0], '--', c='grey')
            ax.plot([0, 0], [-200,200], '--', c='grey')
            
            # # Initialise maximum baseline parameter
            maxSpace = 0
            # Iterate through space telescopes and plot (u,v) coverage
            for i in range(len(spaceTelescopes)):
                for j in range(1, int(stationNo)+1):
                    uv = spaceTelescopes[i].baselines[0][:,j*3-3:j*3]
                    uvLost = spaceTelescopes[i].lostBaselines[0][:,j*3-3:j*3]
                    # Remove zeros indicating no baseline formed
                    uv = uv[~np.all(uv == 0, axis=1)]
                    uvLost = uvLost[~np.all(uvLost == 0, axis=1)]
                    if i == 0 and j == 1:
                        ax.scatter(uv[:,0], uv[:,1], s=m, c='blue', label="Space")
                        if plotLost == 1:
                            ax.scatter(uvLost[:,0], uvLost[:,1], s=m, c='red', \
                                      label="Lost")
                    else:
                        ax.scatter(uv[:,0], uv[:,1], s=m, c='blue')
                    if plotLost == 1:
                        ax.scatter(uvLost[:,0], uvLost[:,1], s=m, c='red') 
                uvMax = np.max(np.array([np.max(spaceTelescopes[i].baselines[0]), \
                                    np.max(spaceTelescopes[i].lostBaselines[0])]))
                if uvMax > maxSpace:
                    maxSpace = uvMax
                    
            # Initialise maximum baseline parameter
            maxGround = 0
            # Iterate through ground telescopes and plot (u,v) coverage
            for i in range(len(groundTelescopes)):
                for j in range(1, int(stationNo)+1):
                    # Vary colour depending on ground or space baseline
                    if j > int(stationNo - len(spaceTelescopes)):
                        colour = 'blue'
                    else:
                        colour = 'orange'                
                    uv = groundTelescopes[i].baselines[0][:,j*3-3:j*3]
                    uvLost = groundTelescopes[i].lostBaselines[0][:,j*3-3:j*3]
                    # Remove zeros indicating no baseline formed
                    uv = uv[~np.all(uv == 0, axis=1)]
                    uvLost = uvLost[~np.all(uvLost == 0, axis=1)]
                    if i == 0 and j == 1:
                        ax.scatter(uv[:,0], uv[:,1], s=m, c=colour, label="Ground")
                        if plotLost == 1:
                            ax.scatter(uvLost[:,0], uvLost[:,1], s=m, c='red') 
                    else:
                        ax.scatter(uv[:,0], uv[:,1], s=m, c=colour)
                        if plotLost == 1:
                            ax.scatter(uvLost[:,0], uvLost[:,1], s=m, c='red') 
                uvMax = np.max(groundTelescopes[i].baselines[0])
                if uvMax > maxGround:
                    maxGround = uvMax
            
            # Configure axes
            axisLimit = 1.3 * np.max(np.array([maxSpace, maxGround]))
            ax.set(xlim=(-axisLimit, axisLimit), ylim=(-axisLimit, axisLimit))
            ax.set_xlabel(r'u [G$ \lambda $]')
            ax.set_ylabel(r'v [G$ \lambda $]')
            ax.set_aspect('equal', 'box')
            ax.legend(loc="upper right")
            ax.xaxis.set_inverted(True)
            plt.show()
            
            #python program to check if a directory exists
            path = "Outputs"
            # Check whether the specified path exists or not
            isExist = os.path.exists(path)
            if not isExist:
               # Create a new directory because it does not exist
               os.makedirs(path)
               print("Creating Outputs folder")
               
            fig.savefig('Outputs/UV.pdf', format='pdf', \
                        bbox_inches='tight')
        else:
            print("Cannot generate (u,v) plot, only one ground or space telescope has been modelled")
    
    else:
        
        # Marker size
        m = 0.1
        # Scale of individual (u,v) plots
        scale = 3
        
        # Calculate number of stations
        stationNo = len(spaceTelescopes) + len(groundTelescopes)

        if stationNo > 1:
    
            fig = plt.figure(figsize=(20,40))
            ax = fig.add_subplot()
            print("Generating all-sky (u,v) coverage plot...")
            
            # Define celestial sphere ranges
            raan = range(0,360,30)
            dec = range(-90,120,30)
            
            # Iterate through right ascension range
            for r in range(len(raan)):
                # Iterate through declination range
                for d in range(len(dec)):
                    
                    # Iterate through space telescopes and plot (u,v) coverage
                    for i in range(len(spaceTelescopes)):
                        for j in range(1, int(stationNo)+1):
                            uv = spaceTelescopes[i].baselines[r*len(dec)+d]\
                                [:,j*3-3:j*3]
                            # Remove zeros indicating no baseline formed
                            uv = uv[~np.all(uv == 0, axis=1)]
                            if i == 0 and j == 1:
                                ax.scatter(uv[:,0]+raan[r]*scale, uv[:,1]+dec[d]*\
                                           scale, s=m, c='blue', label="Space")
                            else:
                                ax.scatter(uv[:,0]+raan[r]*scale, uv[:,1]+dec[d]*\
                                           scale, s=m, c='blue')
                            
                    # Iterate through ground telescopes and plot (u,v) coverage
                    for i in range(len(groundTelescopes)):
                        for j in range(1, int(stationNo)+1):
                            # Vary colour depending on ground or space baseline
                            if j >= int(stationNo - len(spaceTelescopes)):
                                colour = 'blue'
                            else:
                                colour = 'orange'                
                            uv = groundTelescopes[i].baselines[r*len(dec)+d]\
                                [:,j*3-3:j*3]
                            # Remove zeros indicating no baseline formed
                            uv = uv[~np.all(uv == 0, axis=1)]
                            if i == 0 and j == 1:
                                ax.scatter(uv[:,0]+raan[r]*scale, uv[:,1]+dec[d]*\
                                           scale, s=m, c=colour, label="Ground")
                            else:
                                ax.scatter(uv[:,0]+raan[r]*scale, uv[:,1]+dec[d]*\
                                           scale, s=m, c=colour)
            
            # Configure axes
            ax.set_xticks(range(0,scale*360,scale*30), ['0','30','60','90','120',\
                                                  '150','180','210','240','270',\
                                                  '300','330'])
            ax.set_yticks(range(scale*-90,scale*120,scale*30), ['-90','-60','-30',\
                                                  '0','30','60','90'])
            ax.set_xlabel(r'Right Ascension [$ \degree $]')
            ax.set_ylabel(r'Declination [$ \degree $]')
            ax.set_aspect('equal', 'box')
            plt.show()
            
            #python program to check if a directory exists
            path = "Outputs"
            # Check whether the specified path exists or not
            isExist = os.path.exists(path)
            if not isExist:
               # Create a new directory because it does not exist
               os.makedirs(path)
               print("Creating Outputs folder")
               
            fig.savefig('Outputs/UV_AllSky.pdf', format='pdf', \
                         bbox_inches='tight')    
        else:
            print("Cannot generate (u,v) plot, only one ground or space telescope has been modelled")
    
###############################################################################
# AttitudeSphere
###############################################################################

def AttitudeSphere(spaceTelescopes, telescopeSelect=0, azim=45, elev=30, \
                   plotAntenna=1, plotSTR=1, plotRad=1, plotComms=1, plotPanels=1, \
                   plotEarth=1, plotSun=1, plotMoon=1):
    """Plot attitude sphere of spacecraft in the body-fixed frame. Earth, Sun
    and Moon positions depicted. Note. For systems that have different Sun,
    Earth and Moon exclusion angles (e.g. radiator), only the largest is
    plotted.

    :param spaceTelescopes: Array of SpaceTelescope objects, defaults to None
    :type spaceTelescopes: list
    :param telescopeSelect: Index of spaceTelescope array to plot attitude 
        sphere of, defaults to 0
    :type telescopeSelect: int
    :param azim: Plot viewing angle, azimuth in X-Y plane, defaults to 45 degrees
    :type azim: float
    :param elev: Plot viewing angle, elevation in Z plane, defaults to 30 degrees
    :type elev: float
    :param plotAntenna: Plot antenna? Defaults to 1
    :type plotAntenna: bool
    :param plotSTR: Plot star trackers? Defaults to 1
    :type plotSTR: bool
    :param plotRad: Plot radiators? Defaults to 1
    :type plotRad: bool
    :param plotComms: Plot comms systems? Defaults to 1
    :type plotComms: bool
    :param plotPanels: Plot solar panels? Defaults to 1
    :type plotPanels: bool
    :param plotEarth: Plot Earth vector and limb? Defaults to 1
    :type plotEarth: bool
    :param plotSun: Plot Sun vector and limb? Defaults to 1
    :type plotSun: bool
    :param plotMoon: Plot Moon vector and limb? Defaults to 1
    :type plotMoon: bool
    """
    
    if spaceTelescopes:
    
        # Generate figure axis
        fig = plt.figure(figsize=(8,8))
        ax = fig.add_subplot(projection = '3d')
        print("Generating attitude sphere...")
        
        textScale = 1.3
        axisScale = 1.25
        transparency = 0.3
        
        # Extract SpaceTelescope from array and plot attitude sphere
        spaceTelescope = spaceTelescopes[telescopeSelect]
        
        # Plot antenna surface
        if plotAntenna==1:
            vector = spaceTelescope.radioPayloads[0].antBoresight
            [X,Y,Z] = Cone3D([0,0,0], 0.1*vector, 0, 0.15*np.tan(radians(80)), 20)
            ax.plot_surface(X, Y, Z, color='grey', linewidth=0, antialiased=False, \
                            alpha=0.5)
        
        # Plot star trackers
        starTrackers = spaceTelescope.starTrackers
        if spaceTelescope.strModel == 1 and plotSTR == 1 and starTrackers != []:
            for k in range(len(starTrackers)):            
                vector = starTrackers[k].strBoresight
                name = starTrackers[k].name
                # Plot 3D cone showing exclusion angles (only largest exclusion
                # angle is plotted)
                fov = radians(np.max([starTrackers[k].strSunExcl, \
                                      starTrackers[k].strEarthExcl, \
                                          starTrackers[k].strMoonExcl]))            
                ax.plot([0,vector[0]], [0,vector[1]], [0,vector[2]], c='red')
                ax.text(textScale*vector[0], textScale*vector[1], textScale*\
                        vector[2], name, color='red', horizontalalignment='center', verticalalignment='center')
                # Plot 3D cone showing FOV
                if fov > radians(45):
                    scale = (1 - fov/radians(90))*2
                else:
                    scale = 1
                [X,Y,Z] = Cone3D([0,0,0], scale*vector, 0, scale*np.tan(fov), 20)
                ax.plot_surface(X, Y, Z, color='red', linewidth=0, antialiased=False, \
                                alpha=transparency)
            
        # Plot radiator
        radiators = spaceTelescope.radiators
        if spaceTelescope.radModel == 1 and plotRad == 1 and radiators != []:
            for k in range(len(radiators)):            
                vector = radiators[k].radNorm
                name = radiators[k].name
                # Plot 3D cone showing exclusion angles (only largest radiator 
                # exclusion angle is plotted)
                fov = radians(np.max([radiators[k].radSunExcl, \
                                      radiators[k].radEarthExcl, \
                                          radiators[k].radMoonExcl]))
                ax.plot([0,vector[0]], [0,vector[1]], [0,vector[2]], c='red')
                ax.text(textScale*vector[0], textScale*vector[1], textScale*vector[2], \
                        name, color='red', horizontalalignment='center', verticalalignment='center')
                if fov > radians(45):
                    scale = (1 - fov/radians(90))*2
                else:
                    scale = 1
                [X,Y,Z] = Cone3D([0,0,0], scale*vector, 0, scale*np.tan(fov), 20)
                ax.plot_surface(X, Y, Z, color='red', linewidth=0, antialiased=False, \
                                alpha=transparency)
        
        # Plot solar panel directions
        solarPanels = spaceTelescope.solarPanels
        if spaceTelescope.panelModel == 1 and plotPanels == 1 and \
            solarPanels != []:
            for k in range(len(solarPanels)):
                vector = solarPanels[k].panelNorm
                name = solarPanels[k].name
                ax.plot([0,vector[0]], [0,vector[1]], [0,vector[2]], c='red')
                ax.text(textScale*vector[0], textScale*vector[1], textScale*vector[2], \
                        name, color='red', horizontalalignment='center', verticalalignment='center')
                
        # Plot comms system
        commsSystems = spaceTelescope.commsSystems
        if spaceTelescope.commsModel == 1 and plotComms == 1 and commsSystems != []:
            for k in range(len(commsSystems)):
                vector = commsSystems[k].commsNorm
                name = commsSystems[k].name
                fov = radians(commsSystems[k].commsFov)
                ax.plot([0,vector[0]], [0,vector[1]], [0,vector[2]], c='red')
                ax.text(textScale*vector[0], textScale*vector[1], textScale*vector[2], \
                        name, color='red', horizontalalignment='center', verticalalignment='center')
                if fov > radians(45):
                    scale = (1 - fov/radians(90))*2
                else:
                    scale = 1
                [X,Y,Z] = Cone3D([0,0,0], scale*vector, 0, scale*np.tan(fov), 20)
                ax.plot_surface(X, Y, Z, color='red', linewidth=0, antialiased=False, \
                                alpha=transparency)
        
        # Plot initial Earth, Sun and Moon positions
        if plotEarth == 1:
            earthBody = spaceTelescope.earthBody[1,:]
            ax.plot([0,earthBody[0]], [0,earthBody[1]], [0,earthBody[2]], c='limegreen', \
                    label="Earth")
            # Plot 3D cone showing Earth, Sun and Moon limbs
            fov = spaceTelescope.earthLimbAngle[1]
            [X,Y,Z] = Cone3D([0,0,0], earthBody, 0, np.tan(fov), 5)
            ax.plot_surface(X, Y, Z, color='limegreen', linewidth=0, antialiased=False, \
                            alpha=0.1)
        if plotSun == 1:
            sunBody = spaceTelescope.sunBody[1,:]
            ax.plot([0,sunBody[0]], [0,sunBody[1]], [0,sunBody[2]], c='orange', \
                    label="Sun") 
            fov = spaceTelescope.sunLimbAngle[1]
            [X,Y,Z] = Cone3D([0,0,0], sunBody, 0, np.tan(fov), 5)
            ax.plot_surface(X, Y, Z, color='orange', linewidth=0, antialiased=False, \
                            alpha=0.1)
        if plotMoon == 1:
            moonBody = spaceTelescope.moonBody[1,:]
            ax.plot([0,moonBody[0]], [0,moonBody[1]], [0,moonBody[2]], c='blue', \
                    label="Moon")
            fov = spaceTelescope.moonLimbAngle[1]
            [X,Y,Z] = Cone3D([0,0,0], moonBody, 0, np.tan(fov), 5)
            ax.plot_surface(X, Y, Z, color='blue', linewidth=0, antialiased=False, \
                            alpha=0.1)
            
        # Iterate through simulation time and plot Earth, Sun and Moon positions
        for i in range(2,len(spaceTelescope.earthBody)):
            # Plot initial Earth, Sun and Moon positions
            if plotEarth == 1:
                earthBody = spaceTelescope.earthBody[i,:]
                ax.plot([0,earthBody[0]], [0,earthBody[1]], [0,earthBody[2]], c='limegreen')
                # Plot 3D cone showing Earth, Sun and Moon limbs
                fov = spaceTelescope.earthLimbAngle[1]
                [X,Y,Z] = Cone3D([0,0,0], earthBody, 0, np.tan(fov), 5)
                ax.plot_surface(X, Y, Z, color='limegreen', linewidth=0, antialiased=False, \
                                alpha=0.1)
            if plotSun == 1:
                sunBody = spaceTelescope.sunBody[i,:]
                ax.plot([0,sunBody[0]], [0,sunBody[1]], [0,sunBody[2]], c='orange') 
                fov = spaceTelescope.sunLimbAngle[1]
                [X,Y,Z] = Cone3D([0,0,0], sunBody, 0, np.tan(fov), 5)
                ax.plot_surface(X, Y, Z, color='orange', linewidth=0, antialiased=False, \
                                alpha=0.1)
            if plotMoon == 1:
                moonBody = spaceTelescope.moonBody[i,:]
                ax.plot([0,moonBody[0]], [0,moonBody[1]], [0,moonBody[2]], c='blue')
                fov = spaceTelescope.moonLimbAngle[1]
                [X,Y,Z] = Cone3D([0,0,0], moonBody, 0, np.tan(fov), 5)
                ax.plot_surface(X, Y, Z, color='blue', linewidth=0, antialiased=False, \
                                alpha=0.1)
        
        # Plot spacecraft body axis
        ax.plot([0,axisScale], [0,0], [0,0], c='black')
        ax.plot([0,0], [0,axisScale], [0,0], c='black')
        ax.plot([0,0], [0,0], [0,axisScale], c='black')
        ax.text(1.3, 0, 0, "X", color='black', horizontalalignment='center', verticalalignment='center')
        ax.text(0, 1.3, 0, "Y", color='black', horizontalalignment='center', verticalalignment='center')
        ax.text(0, 0, 1.3, "Z", color='black', horizontalalignment='center', verticalalignment='center')
        
        # Configure axes
        axisLimit = 1.3
        ax.set(xlim=(-axisLimit, axisLimit), ylim=(-axisLimit, axisLimit), \
                zlim=(-axisLimit, axisLimit))
        ax.set_xlabel('')
        ax.set_ylabel('')
        ax.set_zlabel('')
        ax.set_xticklabels([])
        ax.set_yticklabels([])
        ax.set_zticklabels([])
        ax.legend(loc="upper right")
        ax.set_aspect('equal', 'box')  
        ax.view_init(elev, azim)
        plt.show()
        
        #python program to check if a directory exists
        path = "Outputs"
        # Check whether the specified path exists or not
        isExist = os.path.exists(path)
        if not isExist:
           # Create a new directory because it does not exist
           os.makedirs(path)
           print("Creating Outputs folder")
           
        fig.savefig('Outputs/AttitudeSphere.pdf', format='pdf', \
                    bbox_inches='tight')
    else:
        print("Cannot generate Attitude Sphere plot, no space telescopes are modelled")

###############################################################################
# SolarPanelIncidence
###############################################################################

def SolarPanelIncidence(spaceTelescopes, simTime, telescopeSelect=0):
    """Plot angle of Sun incidence on the solar panels, measured with respect to
    the panel normal. 

    :param spaceTelescopes: Array of SpaceTelescope objects, defaults to None
    :type spaceTelescopes: list
    :param simTime: Timeseries of simulation time, defaults to None
    :type simTime: list
    :param telescopeSelect: Index of spaceTelescope array to plot solar panel 
        incidence of, defaults to 0
    :type telescopeSelect: int
    """

    if spaceTelescopes:
        # Extract SpaceTelescope object
        spaceTelescope = spaceTelescopes[telescopeSelect]
        
        if spaceTelescope.panelModel == 1:
            
            fig = plt.figure(figsize=(8,8))
            ax = fig.add_subplot()
            print("Generating Solar Panel Angle plot...")
            
            # Extract time
            time = simTime.time.value[1:]
            for j in range(len(time)):
                time[j] = time[j][11:-4]
                
            # Iterate through solar panels and plot angle
            solarPanels = spaceTelescope.solarPanels
            for i in range(len(solarPanels)):
                name = solarPanels[i].name
                betaAngle = solarPanels[i].betaAngle
                ax.plot(time, betaAngle[1:], label=name)
                
            # Configure axes
            ax.set_xlabel('Time')
            ax.set_ylabel(r'Incidence Angle [$ \degree $]')
            ax.set_xticks(time[0::int(np.floor(len(time)/10))])
            ax.legend(loc="upper right")
            plt.xticks(rotation=90)
            plt.show()
            
            #python program to check if a directory exists
            path = "Outputs"
            # Check whether the specified path exists or not
            isExist = os.path.exists(path)
            if not isExist:
               # Create a new directory because it does not exist
               os.makedirs(path)
               print("Creating Outputs folder")
               
            fig.savefig('Outputs/SolarPanelIncidence.pdf', format='pdf', \
                        bbox_inches='tight')
        else:
            print("Solar panel modelling not turned on for selected space telescope")
    else:
        print("Cannot generate Solar Panel Incidence plot, no space telescopes are modelled")

###############################################################################
# GroundStationElevation
###############################################################################

def GroundStationElevation(spaceTelescopes, groundStations, simTime, \
                           telescopeSelect=0):
    """Plot elevation angle of selected space telescope at each ground station
    as a function of the simulation time. 

    :param spaceTelescopes: Array of SpaceTelescope objects, defaults to None
    :type spaceTelescopes: list
    :param groundStations: Array of GroundStation objects, defaults to None
    :type groundStations: list
    :param simTime: Timeseries of simulation time, defaults to None
    :type simTime: list
    :param telescopeSelect: Index of spaceTelescope array to plot ground 
        station elevation of, defaults to 0
    :type telescopeSelect: int
    """
    
    if spaceTelescopes:
        if len(groundStations) == 0:
            print("No ground stations are modelled")
        else:
            
            fig = plt.figure(figsize=(8,8))
            ax = fig.add_subplot()
            print("Generating Ground Station Access plot...")
            
            # Extract time
            time = simTime.time.value[1:]
            
            # Iterate through solar panels and plot angle
            # solarPanels = spaceTelescope.solarPanels
            for i in range(len(groundStations)):
                name = groundStations[i].name
                elevation = groundStations[i].satElev[1:,telescopeSelect]
                ax.plot(time, elevation, label=name)
                
            # Configure axes
            ylabel = spaceTelescopes[telescopeSelect].name + r" Elevation Angle [$ \
                \degree $]"
            ax.set(ylim=(0, 90))
            ax.set_xlabel('Time')
            ax.set_ylabel(ylabel)
            ax.set_xticks(time[0::int(np.floor(len(time)/10))])
            ax.legend(loc="upper right")
            plt.xticks(rotation=90)
            plt.show()
            
            #python program to check if a directory exists
            path = "Outputs"
            # Check whether the specified path exists or not
            isExist = os.path.exists(path)
            if not isExist:
               # Create a new directory because it does not exist
               os.makedirs(path)
               print("Creating Outputs folder")
            
            fig.savefig('Outputs/GroundStationAccess.pdf', format='pdf',\
                        bbox_inches='tight')
    else:
        print("Cannot generate Ground Station Elevation plot, no space telescopes are modelled")
        
###############################################################################
# Elevation Earth Map
###############################################################################
        
def ElevationEarthMap(spaceTelescopes, groundStations, telescopeSelect=0):
    """Plot Earth map showing spacecraft ground track and visibility to ground
    stations. 

    :param spaceTelescopes: Array of SpaceTelescope objects, defaults to None
    :type spaceTelescopes: list
    :param groundStations: Array of GroundStation objects, defaults to None
    :type groundStations: list
    :param telescopeSelect: Index of spaceTelescope array to plot ground 
        station elevation of, defaults to 0
    :type telescopeSelect: int
    """
    
    url = "https://naciscdn.org/naturalearth/110m/cultural/ne_110m_admin_0_countries.zip"
    gdf = geopandas.read_file(url)
    spacecraft = spaceTelescopes[telescopeSelect]
    colours = ['g', 'b', 'r', 'c', 'y', 'm', 'orange', 'limegreen', 'lightsteelblue', \
               'gold', 'brown', 'indigo', 'violet', 'whitesmoke']
    # Creating axes and plotting world map
    fig, ax = plt.subplots(figsize=(16, 10))
    gdf.plot(color="lightgrey", ax=ax)
    
    # Iterate through ground stations and plot
    for i in range(0,len(groundStations)):
        gs_ecef = groundStations[i].ecefPosition / 1000
        
        r_sc = np.linalg.norm(spacecraft.eciPosition[0,:].value)
        r_gs = np.linalg.norm(gs_ecef*1000)
        min_el = groundStations[i].minEl
        beta = np.arcsin(r_gs*np.sin(np.radians(min_el+90)) / r_sc)
        radius = np.degrees(np.pi - np.radians(min_el+90) - beta)
        
        gs_lla = ecef_to_lla(gs_ecef)
        lon_0 = np.radians(gs_lla[1])
        lat_0 = np.radians(gs_lla[0])
                
        if gs_ecef[0] < 0:
            lon_0 = -np.pi - lon_0
        if lon_0 < -np.pi:
            lon_0 = lon_0 + 2*np.pi
        rad = np.radians(radius)
        lat=[]
        lon=[]
        for beta in np.linspace(0,np.pi*2,200):
           lat_new = ((np.pi/2) - np.arccos(np.cos((np.pi/2)-lat_0)*np.cos(rad) + np.sin((np.pi/2)-lat_0)*np.sin(rad)*np.cos(beta)))
           if beta == 0 or beta == 2*np.pi:
               lon_delta = 0
           if beta < np.pi and beta > 0:
               lon_delta = np.arccos((np.cos(rad) - np.cos((np.pi/2)-lat_0)*np.cos((np.pi/2)-lat_new)) / (np.sin((np.pi/2)-lat_0)*np.sin((np.pi/2)-lat_new)))
           if beta >= np.pi and beta < 2*np.pi:
               lon_delta = -np.arccos((np.cos(rad) - np.cos((np.pi/2)-lat_0)*np.cos((np.pi/2)-lat_new)) / (np.sin((np.pi/2)-lat_0)*np.sin((np.pi/2)-lat_new)))
           lat_new = np.degrees(lat_new)
           lon_new = np.degrees(lon_0 + lon_delta)
           lat.append(lat_new)
           lon.append(lon_new)
        
        if (np.abs(lat_0) + rad) > np.pi/2:
            lon2 = [0,180,180,0,-180,-180,0]
            offset = -0.1
            lat2 = [np.max(np.abs(lat))+offset,np.max(np.abs(lat))+offset,90,90,90,np.max(np.abs(lat))+offset,np.max(np.abs(lat))+offset]
            lat2 = [i * np.sign(lat_0) for i in lat2]
            
            polygon_geom2 = Polygon(zip(lon2, lat2))
            pol2 = geopandas.GeoDataFrame(index=[0], crs='epsg:4326', geometry=[polygon_geom2])
            pol2.plot(ax=ax, alpha=0.5, fc=colours[i], ec='none')  
        
        if np.abs(lon_0) + rad > np.radians(170):
            lon3 = [i - np.sign(lon_0)*360 for i in lon]
            lat3 = lat
            polygon_geom3 = Polygon(zip(lon3, lat3))
            pol3 = geopandas.GeoDataFrame(index=[0], crs='epsg:4326', geometry=[polygon_geom3])
            pol3.plot(ax=ax, alpha=0.5, fc=colours[i], ec='none')
            
        plt.scatter(np.degrees(lon_0), np.degrees(lat_0), s=100, c=colours[i], alpha=0.6, label=groundStations[i].name)
        polygon_geom = Polygon(zip(lon, lat))
        pol = geopandas.GeoDataFrame(index=[0], crs='epsg:4326', geometry=[polygon_geom])
        pol.plot(ax=ax, alpha=0.5, fc=colours[i], ec='none') 
        
        # Extract spacecraft positions when visible to ground stations and plot
        latlon = np.zeros((len(groundStations[i].satElev),2))
        for j in range(0,len(groundStations[i].satElev)):
            sc_ecef = spacecraft.ecefPosition[j,:].value
            sc_lla = ecef_to_lla(sc_ecef)
            sc_x = sc_lla[1]
            sc_y = sc_lla[0]
            if sc_ecef[0] < 0:
                sc_x = -180 - sc_lla[1]
            if sc_x < -180:
                sc_x = 360 + sc_x
            latlon[j,:] = [sc_x,sc_y]
        ax.scatter(latlon[:,0], latlon[:,1], s=10, c='k')
    
    # Creating axis limits and title
    plt.xlim([-180, 180])
    plt.ylim([-90, 90])
    
    #plt.title("Tourist arrivals from main countries to Sri Lanka\n  Year : 2021")
    plt.xlabel(r"Longitude [$\degree$]")
    plt.ylabel(r"Latitude [$\degree$]")
    ax.set_aspect('equal')
    ax.legend(loc="upper right")
    ax.set_title('Ground station visibility (minimum elevation = ' + str(min_el) + r'$\degree$)')
    plt.show()

    #python program to check if a directory exists
    path = "Outputs"
    # Check whether the specified path exists or not
    isExist = os.path.exists(path)
    if not isExist:
       # Create a new directory because it does not exist
       os.makedirs(path)
       print("Creating Outputs folder")
       
    fig.savefig('Outputs/GroundStationContact.pdf', dpi=600, format='pdf', \
                bbox_inches='tight')

###############################################################################
# Miscellaneous
###############################################################################
    
def Cone3D(p0, p1, R0, R1, n):
    """Plot 3D cone. Used in AttitudeSphere figure.
 
    :param p0: Unit vector of vertex position, defaults to None
    :type p0: list
    :param p1: Unit vector of base position, defaults to None
    :type p1: list
    :param R0: Radius at vertex, defaults to None
    :type R0: float
    :param R1: Radius at base, defaults to None
    :type R1: float
    :param n: Number of radial steps used in cone surface generation, defaults 
        to None
    :type n: int
    """
    # vector in direction of axis
    v = p1 - p0
    # find magnitude of vector
    mag = np.linalg.norm(v)
    # unit vector in direction of axis
    v = v / mag
    # make some vector not in the same direction as v
    not_v = np.array([1, 1, 0])
    if (v == not_v).all():
        not_v = np.array([0, 1, 0])
    # make vector perpendicular to v
    n1 = np.cross(v, not_v)
    # print n1,'\t',norm(n1)
    # normalize n1
    n1 /= np.linalg.norm(n1)
    # make unit vector perpendicular to v and n1
    n2 = np.cross(v, n1)
    # surface ranges over t from 0 to length of axis and 0 to 2*pi
    t = np.linspace(0, mag, n)
    theta = np.linspace(0, 2 * np.pi, n)
    # use meshgrid to make 2d arrays
    t, theta = np.meshgrid(t, theta)
    R = np.linspace(R0, R1, n)
    # generate coordinates for surface
    X, Y, Z = [p0[i] + v[i] * t + R *
               np.sin(theta) * n1[i] + R * np.cos(theta) * n2[i] for i in \
              [0, 1, 2]]
    return X, Y, Z

def ecef_to_lla(ecef_pos):
    """Function to convert Earth-centered Earth-fixed coordinates to latitude, 
    longitude and altitude. Equation from Vallado Algorithm 12.
    
    :param ecef_pos: Position vector in ECEF, defaults to None
    :type ecef_pos: np.array
    :return: Array of latitude, longitude and altitude
    :rtype: list
    """
    
    r_delta_sat = np.sqrt(ecef_pos[0]**2 + ecef_pos[1]**2)
    # Use WGS84 reference ellipsoid for transformation
    a = 6378.137
    b = 6356.752314

    e = np.sqrt(1 - b**2 / a**2)
    
    alpha = np.arcsin(ecef_pos[1] / r_delta_sat)
    lam = alpha
    delta = np.arcsin(ecef_pos[2] / np.linalg.norm(ecef_pos))
    
    # Iterate to calculate latitude
    phi = delta
    phi_old = 0
    tol = 0.01  # tolerance for iteration
    while phi - phi_old > tol:
        c = a / np.sqrt(1 - e**2*np.sin(phi)**2)
        phi_old = phi
        phi = np.arctan((ecef_pos[2] + c*e**2*np.sin(phi)) / r_delta_sat)
        
    # Calculate altitude
    c = a / np.sqrt(1 - e**2*np.sin(phi)**2)
    if (90 - np.abs(np.degrees(lam))) < 1:
        s = c * (1-e**2)
        h = (ecef_pos[2] / np.sin(phi)) - s
    else:
        h = (r_delta_sat / np.cos(phi)) - c
    
    return([np.degrees(phi),np.degrees(lam),h])



"""
Library for simulating space-based VLBI missions (spacevlbi)

Copyright 2024 Ben Hudson

Licensed under the GNU General Public License, Version 3.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.gnu.org/licenses/gpl-3.0.en.html

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# Station.py
#
# This file contains the definitions of three classes used throughout spacevlbi:
# SpaceTelescope, GroundTelescope and GroundStation. Also included are classes
# for defining spacecraft components that can be modelled within the tool.
#
# @author: BenHudson - 22/09/2024

from astropy.coordinates import SkyCoord,GCRS,ITRS
from astropy import constants as const
import numpy as np
from astropy import units as u
from poliastro.twobody import Orbit
from poliastro.bodies import Earth

###############################################################################
#   SpaceTelescope
###############################################################################

class SpaceTelescope:
    """Class to model a space telescope within the spacevlbi package.

    :param name: Space telescope name, defaults to None
    :type name: str 
    :param mass: Spacecraft mass in kg, defaults to None
    :type mass: float
    :param areaDrag: Spacecraft drag area in m^2, defaults to None
    :type areaDrag: float
    :param areaSolar: Spacecraft solar area in m^2, defaults to None
    :type areaSOlar: float
    :param cD: Drag coefficient, defaults to None
    :type cD: float
    :param cR: Solar reflectivity, defaults to None
    :type cR: float
    :param initTime: Simulation start time, defaults to None
    :type initTime: str
    :param a: Orbit semi-major axis in km, defaults to None
    :type a: float
    :param ecc: Orbit eccentricity, defaults to None
    :type ecc: float
    :param inc: Orbit inclination in degrees, defaults to None
    :type inc: float
    :param aop: Orbit argument of perigee in degrees, defaults to None
    :type aop: float
    :param ta: Orbit true anomaly in degrees, defaults to None
    :type ta: float
    :param pointingVector: Body-fixed vector to point at target source 
        (should be same as RadioPayload.antennaBoresight), defaults to None
    :type pointingVector: numpy.ndarray
    :param constraintVector: Body-fixed vector to point in direction 
        perpendicular to target source. Must be perpendicular to pointingVector,
        defaults to None
    :type constraintVector: numpy.ndarray
    :param rollAngle: Roll angle of space telescope in degrees about the 
        pointingVector. Measured from the plane containing the celestial 
        north pole and the target source direction. Clockwise direction is 
        positive when viewing along the pointingVector direction. Roll 
        angle control is achieved by defining a list of the following 
        format: [transition 1 time in sec, Roll angle 1 in degrees, 
        transition 2 time in sec, roll angle 2, etc.], defaults to None
    :type rollAngle: list
    :param radioPayloads: Array of RadioPayload objects, defaults to None
    :type radioPayloads: list
    :param starTrackers: Array of StarTracker objects, defaults to None
    :type starTrackers: list
    :param reqStarTrackers: Number of star trackers that must not be
        blinded to conduct an observation, defaults to None
    :type reqStarTrackers: int
    :param radiators: Array of Radiator objects, defaults to None
    :type radiators: list
    :param commsSystems: Array of CommsSystem objects, defaults to None
    :type commsSystems: list
    :param solarPanels: Array of SolarPanel objects, defaults to None
    :type solarPanels: list
    :param strModel: Flag indicating whether StarTrackers should be
        modelled, defaults to None
    :type strModel: bool
    :param radModel: Flag indicating whether Radiators should be
        modelled, defaults to None
    :type radModel: bool
    :param commsModel: Flag indicating whether CommsSystems should be
        modelled, defaults to None
    :type commsModel: bool
    :param panelModel: Flag indicating whether SolarPanels should be
        modelled, defaults to None
    :type panelModel: bool
    :param gravityJ2: Model Earth gravity harmonic J2? Defaults to None
    :type gravityJ2: BOOL
    :param gravityJ3: Model Earth gravity harmonic J3? Defaults to None
    :type gravityJ3: BOOL
    :param atmosDrag: Model atmospheric drag? Defaults to None
    :type atmosDrag: BOOL
    :param solarPress: Model solar radiation pressure? Defaults to None
    :type solarPress: BOOL
    :param solarFlux: Solar flux in W/m^2, defaults to None
    :type solarFlux: float
    :param gravityLuniSolar: Model luni-solar point model gravity? Defaults to None
    :type gravityLuniSolar: BOOL
    :return: SpaceTelescope object
    :rtype: SpaceTelescope
    """
        
    def __init__(self, name, mass, areaDrag, areaSolar, cD, cR, initTime, a, \
                 ecc, inc, ra, aop, ta,  pointingVector, constraintVector, \
                 rollAngle, radioPayloads, starTrackers, reqStarTrackers, \
                 radiators, commsSystems, solarPanels, strModel, radModel, 
                 commsModel, panelModel, gravityJ2, gravityJ3, atmosDrag, \
                 solarPress, solarFlux, gravityLuniSolar):
               
        self.name = name
        self.pointingVector = np.array([pointingVector])
        self.constraintVector = np.array([constraintVector])
        self.rollAngle = rollAngle
        self.a = a << u.km
        self.ecc = ecc << u.one
        self.inc = inc << u.deg
        self.ra = ra << u.deg
        self.aop = aop << u.deg
        self.ta = ta << u.deg
        self.orbit = Orbit.from_classical(Earth, self.a, self.ecc, self.inc, \
                      self.ra, self.aop, self.ta, initTime)
        self.orbitLast = self.orbit
        self.eciPosition = np.array([self.orbit.r*1000])<< u.m
        self.eciVelocity = np.array([self.orbit.v*1000])<< (u.m / u.s)
        eci =SkyCoord(x=self.eciPosition[0,0],y=self.eciPosition[0,1],\
                      z=self.eciPosition[0,2], unit='m',frame='gcrs',\
                      representation_type='cartesian', obstime = initTime)
        ecef = (eci.transform_to(ITRS)).cartesian
        self.ecefPosition = np.array([[ecef.x.value, ecef.y.value, \
                      ecef.z.value]]) << u.m
        self.earthLimbAngle = np.array([0])
        self.sunLimbAngle = np.array([0])
        self.moonLimbAngle = np.array([0])
        self.eclipse = np.array([0])
        self.mass = mass
        self.areaDrag = areaDrag
        self.areaSolar = areaSolar
        self.cD = cD
        self.cR = cR
        self.radioPayloads = radioPayloads
        self.strModel = strModel
        self.starTrackers = starTrackers
        self.reqStarTrackers = reqStarTrackers
        self.panelModel = panelModel
        self.gravityJ2 = gravityJ2
        self.gravityJ3 = gravityJ3
        self.atmosDrag = atmosDrag
        self.solarPress = solarPress
        self.solarFlux = solarFlux
        self.gravityLuniSolar = gravityLuniSolar
        self.solarPanels = solarPanels
        self.radModel = radModel
        self.radiators = radiators
        self.commsModel = commsModel
        self.commsSystems = commsSystems
        self.sunInertial = np.array([[0,0,0]])  # Sun vector in inertial frame
        self.moonInertial = np.array([[0,0,0]])  # Moon vector in inertial frame
        self.sourceInertial = np.array([[0,0,0]])  # Earth vector in inertial frame
        self.rSun = np.array([[0,0,0]])  # Sat - Sun vector
        self.rMoon = np.array([[0,0,0]])  # Sat - Moon vector
        self.sunBody = np.array([[0,0,0]])  # Sun vector in body-fixed frame
        self.moonBody = np.array([[0,0,0]])  # Moon vector in body-fixed frame
        self.earthBody = np.array([[0,0,0]])  # Earth vector in body-fixed frame
        self.baselines = []  # VLBI baselines
        self.lostBaselines = []  # VLBI baselines lost due to functional constraints
        # Flag indicating whether the source is in view of the spacecraft
        self.sourceVisibility = np.array([0])
        
        
###############################################################################
#   GroundTelescope
###############################################################################

class GroundTelescope:
    """Class to model a ground telescope within the spacevlbi package.
    
    :param name: Ground telescope name, defaults to None
    :type name: str 
    :param diameter: Antenna diameter in metres, defaults to None
    :type diameter: float
    :param apertureEff: Aperture efficiency due to surface design and losses
        from surface deformations, defaults to None
    :type apertureEff: float
    :param sysTemp: System temperature in Kelvin, defaults to None
    :type sysTemp: float
    :param ecefPosition: Ground telescope position in Earth Centered, Earth
        Fixed (ECEF) frame in km, defaults to None
    :type ecefPosition: numpy.ndarray
    :param minEl: Minimum elevation at which observation can be performed
        in degrees, defaults to None
    :type minEl: float
    :param initTime: Simulation start time, defaults to None
    :type initTime: str
    :return: GroundTelescope object
    :rtype: GroundTelescope
    """
    
    def __init__(self, name, diameter, apertureEff, sysTemp, \
               ecefPosition, minEl, initTime):

        self.name = name
        self.diameter = diameter
        self.sysTemp = sysTemp
        self.apertureEff = apertureEff
        self.SEFD = (2*const.k_B*sysTemp)/(self.apertureEff * np.pi *\
            (diameter/2)**2)/(10**-26);  # Jy
        self.ecefPosition = np.array([ecefPosition*1000 << u.m])
        self.minEl = minEl
        # Convert ECEF to ECI positions
        ecef =SkyCoord(x=ecefPosition[0],y=ecefPosition[1],z=ecefPosition[2],\
                       unit='m',frame='itrs',representation_type='cartesian',\
                           obstime = initTime)
        eci = (ecef.transform_to(GCRS)).cartesian       
        self.eciPosition = np.array([[eci.x.value, eci.y.value, eci.z.value]])\
                       << u.m
        self.baselines = []  # VLBI baselines
        self.lostBaselines = []  # VLBI baselines lost due to functional constraints
        # Elevation of source from ground telescope in topocentric frame
        self.elevation = np.array([0])
        self.elevationFlag = np.array([0])


###############################################################################
#   GroundStation
###############################################################################

class GroundStation:
    """Class to model a ground station within the spacevlbi package.
    
    :param name: Ground telescope name, defaults to None
    :type name: str 
    :param ecefPosition: Ground station position in Earth Centered, Earth
        Fixed (ECEF) frame in km, defaults to None
    :type ecefPosition: numpy.ndarray
    :param minEl: Minimum elevation at which link with spacecraft can be 
        achieved in degrees, defaults to None
    :type minEl: float
    :param initTime: Simulation start time, defaults to None
    :type initTime: str
    :return: GroundStation object
    :rtype: GroundStation
    """
    
    def __init__(self, name, ecefPosition, minEl, initTime):
        
        self.name = name
        self.ecefPosition = ecefPosition*1000
        self.minEl = minEl
        # Convert ECEF to ECI positions
        ecef =SkyCoord(x=ecefPosition[0],y=ecefPosition[1],z=ecefPosition[2],\
                       unit='m',frame='itrs',representation_type='cartesian',\
                           obstime = initTime)
        eci = (ecef.transform_to(GCRS)).cartesian       
        self.eciPosition = np.array([eci.x.value, eci.y.value, eci.z.value])*1000 \
                        << u.m
        # Range of spacecraft from ground station
        self.satRange = np.array([0]) << u.m
        # Elevation of spacecraft from ground station
        self.satElev = np.array([0])
        
        
###############################################################################
#   Spacecraft Components
###############################################################################

class RadioPayload:
    """Class to model a radio astronomy payload onboard a SpaceTelescope.
    
    :param name: Radio payload name, defaults to None
    :type name: str 
    :param diameter: Antenna diameter in metres, defaults to None
    :type diameter: float
    :param apertureEff: Aperture efficiency due to surface design and losses
        from surface deformations, defaults to None
    :type apertureEff: float
    :param sysTemp: System temperature in Kelvin, defaults to None
    :type sysTemp: float        
    :param antBoresight: Antenna boresight in body-fixed frame, defaults to
        None
    :type antBoresight: numpy.ndarray
    :param antSunExcl: Exclusion angle between antenna boresight and the 
        Sun's limb in degrees, defaults to None
    :type antSunExcl: float
    :param antEarthExcl: Exclusion angle between antenna boresight and the 
        Earth's limb in degrees, defaults to None
    :type antEarthExcl: float
    :param antMoonExcl: Exclusion angle between antenna boresight and the 
        Moon's limb in degrees, defaults to None
    :type antMoonExcl: float
    :return: RadioPayload object
    :rtype: RadioPayload
    """
    
    def __init__(self, name, diameter, apertureEff, sysTemp,\
                 antBoresight, antSunExcl, antEarthExcl, \
                 antMoonExcl):
 
        self.name = name
        self.diameter = diameter
        self.sysTemp = sysTemp
        self.apertureEff = apertureEff
        self.SEFD = (2*const.k_B*sysTemp)/(self.apertureEff * np.pi *\
            (diameter/2)**2)/(10**-26)  # Jy
        self.antBoresight = antBoresight
        self.antSunExcl = antSunExcl
        self.antEarthExcl = antEarthExcl
        self.antMoonExcl = antMoonExcl
        # Antenna normal in inertial attitude system
        self.antennaInertial = np.array([0,0,0])
        # Havw the Sun/Earth/Moon exclusion angles been violated?
        self.sunFlag = np.array([0])
        self.earthFlag = np.array([0])
        self.moonFlag = np.array([0])
        

class StarTracker:
    """Class to model a star tracker onboard a SpaceTelescope.
    
    :param name: Star tracker name, defaults to None
    :type name: str       
    :param strBoresight: Star tracker boresight in body-fixed frame, 
        defaults to None
    :type strBoresight: numpy.ndarray
    :param strSunExcl: Exclusion angle between star tracker boresight and 
        the Sun's limb in degrees, defaults to None
    :type strSunExcl: float
    :param strEarthExcl: Exclusion angle between star tracker boresight and
        the Earth's limb in degrees, defaults to None
    :type strEarthExcl: float
    :param strMoonExcl: Exclusion angle between star tracker boresight and 
        the Moon's limb in degrees, defaults to None
    :type strMoonExcl: float
    :return: StarTracker object
    :rtype: StarTracker
    """
    
    def __init__(self, name, strBoresight, strSunExcl, strEarthExcl, \
                 strMoonExcl):
        
        self.name = name
        self.strBoresight = strBoresight
        self.strSunExcl = strSunExcl
        self.strEarthExcl = strEarthExcl
        self.strMoonExcl = strMoonExcl
        # Flag indicating whether star tracker is blinded (1==blinded)
        self.strBlindFlag = np.array([0])
        # StarTracker normal in inertial attitude system
        self.strInertial = np.array([[0,0,0]])
        
    
class Radiator:
    """Class to model a radiator onboard a SpaceTelescope.
    
    :param name: Radiator name, defaults to None
    :type name: str       
    :param radNorm: Radiator normal vector in body-fixed frame, 
        defaults to None
    :type radNorm: numpy.ndarray
    :param radSunExcl: Exclusion angle between radiator normal and 
        the Sun's limb in degrees, defaults to None
    :type radSunExcl: float
    :param radEarthExcl: Exclusion angle between radiator normal and
        the Earth's limb in degrees, defaults to None
    :type radEarthExcl: float
    :param radMoonExcl: Exclusion angle between radiator normal and 
        the Moon's limb in degrees, defaults to None
    :type radMoonExcl: float
    :return: Radiator object
    :rtype: Radiator
    """
    
    def __init__(self, name, radNorm, radSunExcl, radEarthExcl, \
                 radMoonExcl):
        
        self.name = name
        self.radNorm = radNorm
        self.radSunExcl = radSunExcl
        self.radEarthExcl = radEarthExcl
        self.radMoonExcl = radMoonExcl
        # Radiator normal in inertial attitude system
        self.radInertial = np.array([[0,0,0]])
        # Flag indicating whether radiator is blinded (1==blinded)
        self.radBlindFlag = np.array([0])

        
class SolarPanel:
    """Class to model a solar panel onboard a SpaceTelescope.
    
    :param name: Solar panel name, defaults to None
    :type name: str       
    :param panelNorm: Solar panel normal vector in body-fixed frame, 
        defaults to None
    :type panelNorm: numpy.ndarray
    :return: SolarPanel object
    :rtype: SolarPanel
    """
    
    def __init__(self, name, panelNorm):
        
        self.name = name
        self.panelNorm = panelNorm
        # Panel normal in inertial attitude system
        self.solarPanelInertial = np.array([[0,0,0]])
        # Angle between panel normal and solar direction
        self.betaAngle = np.array([0])
        
        
class CommsSystem:
    """Class to model a communications system onboard a SpaceTelescope.
    
    :param name: Comms system name, defaults to None
    :type name: str       
    :param commsNorm: Comms system normal vector in body-fixed frame, 
        defaults to None
    :param commsFov: Defines the field of view of the comms system within 
        which, if a ground station is insight, a link session can be
        maintained. Could be the half angle beamwidth of a radio or the gimbal
        limit of an optical terminal, etc., defaults to None
    :type commsFov: float
    :param groundReqObs: Ground station contact required for observations? 
        Defaults to None
    :type groundReqObs: bool
    :return: CommsSystem object
    :rtype: CommsSystem
    """
    
    def __init__(self, name, commsNorm, commsFov, groundReqObs):
        
        self.name = name
        self.commsNorm = commsNorm
        self.commsFov = commsFov
        self.groundReqObs = groundReqObs
        # Comms system normal in inertial attitude system
        self.commsInertial = np.array([[0,0,0]])
        # Angle between comms system normal and ground station direction
        self.commsGroundAngle = np.array([0])
        # Array of flags indicating which ground stations are insight
        self.groundStationInsight = np.array([0])



# -*- coding: utf-8 -*-
#
# __init__.py
#
#@author: BenHudson - 28/07/2024

from spacevlbi.Attitude import *
from spacevlbi.Constraints import *
from spacevlbi.Figures import *
from spacevlbi.Observation import *
from spacevlbi.Orbit import *
from spacevlbi.Station import *
from spacevlbi.TimeLoop import *
from spacevlbi.Optimisation import *


"""
Library for simulating space-based VLBI missions (spacevlbi)

Copyright 2024 Ben Hudson

Licensed under the GNU General Public License, Version 3.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.gnu.org/licenses/gpl-3.0.en.html

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# ExampleSetup.py
#
# An example implementation of the spacevlbi package to demonstrate how the
# functions can be executed to perform simulation of a space-based VLBI mission.
# The example given is a simplified representation of the Black Hole Explorer
# (BHEX) mission - https://www.blackholeexplorer.org/

# @author: BenHudson - 22/09/2024

from spacevlbi import Station
from spacevlbi.TimeLoop import TimeLoop
from spacevlbi import Figures
from spacevlbi.Optimisation import Optimisation
import numpy as np
from poliastro.util import Time
from ExampleSpaceTelescope import BaselineBHEX
import matplotlib

# Plot configurations
matplotlib.rcParams['mathtext.fontset'] = 'stix'
matplotlib.rcParams['font.family'] = 'STIXGeneral'
matplotlib.rcParams.update({'font.size': 20})

###############################################################################
# Simulation Time Definition
###############################################################################

initTime = Time("2025-01-01 00:00", scale="utc")  # start time of simulation
timeStep = 100  # simulation time step, sec
simLength = 86400  # length of simulation, sec

###############################################################################
# Observation Parameters
###############################################################################

obsFreq = 320e9  # frequency observations will be conducted at, Hz
# Calculate (u,v) coverage for full celestial sphere? NOTE. Functional
# constraints cannot be modelled in all-sky mode.
allsky = 0

# M87*
sourceRa = 187.705930  # target source right ascension, deg
sourceDec = 12.391123  # target source declination, deg

# Sgr A*
#sourceRa = 266.25  # target source right ascension, deg
#sourceDec = -29.0078 # target source declination, deg

intTime = 0  # integration time, sec
dutyCycle = 0  # time between the start of one scan and the next, sec
bandwidth = 32e9  # bandwidth of observations, sec

###############################################################################
# Ground Telescope Definition
###############################################################################

# See GroundTelescope class for definition of input parameters
elev = 15
ALMA = Station.GroundTelescope("ALMA", 73, 0.68, 76, \
                np.array([2225.0613,-5440.0617,-2481.6812]), elev, initTime)
IRAM = Station.GroundTelescope("IRAM", 30, 0.47, 226, \
                np.array([5088.9678,-301.6812,3825.0122]), elev, initTime)
APEX = Station.GroundTelescope("APEX", 12, 0.61, 118, \
                np.array([2225.0395, -5441.1976, -2479.3034]), elev, initTime)
JCMT = Station.GroundTelescope("JCMT", 15, 0.52, 345, \
                np.array([-5464.5847, -2493.0012, 2150.6540]), elev, initTime)
LMT = Station.GroundTelescope("LMT", 32.5, 0.28, 371, \
                np.array([-768.7156, -5988.5071, 2063.3549]), elev, initTime)
SMA = Station.GroundTelescope("SMA", 14.7, 0.75, 285, \
                np.array([-5464.5555, -2492.9280, 2150.7972]), elev, initTime)
SMT = Station.GroundTelescope("SMT", 10, 0.60, 291, \
                np.array([-1828.7962, -5054.4068, 3427.8652]), elev, initTime)
SPT = Station.GroundTelescope("SPT", 6, 0.60, 118, \
                np.array([0.8098, -0.8169, -6359.5687]), elev, initTime)
NOEMA = Station.GroundTelescope("NOEMA", 52,   0.50, 270,\
                np.array([4524.0004, 468.0421, 4460.5098]), elev, initTime)
HAY = Station.GroundTelescope("HAY", 52, 0.50, 270, \
                np.array([1492.341, -4457.234,  4296.933]), elev, initTime)
PV = Station.GroundTelescope("PV", 15, 0.47, 226, \
                np.array([5088.9678, -301.6812,  3825.012]), elev, initTime)

###############################################################################
# Spacecraft Definition
###############################################################################

# See ExampleSpaceTelescope.py script for spacecraft definition and the 
# SpaceTelescope class for definition of input parameters
sc1 = BaselineBHEX(initTime)

###############################################################################
# Ground Station Definition
###############################################################################

# See GroundStation class for definition of input parameters
Svalbard = Station.GroundStation("Svalbard", \
                np.array([1258.4, 346.3, 6222.2]), 5, initTime);
Haleakala = Station.GroundStation("Haleakala", \
                np.array([-5463.394,-2403.165, 2241.185]), 15, initTime);
Lasilla = Station.GroundStation("La Silla", \
                np.array([1837.134,-5257.272,-3098.702]), 15, initTime);
Achaea = Station.GroundStation("Achaea", \
                np.array([4660.348,1901.702,3904.176]), 15, initTime);
Perth = Station.GroundStation("Perth", \
                np.array([-2384.681,4860.089,-3361.149]), 15, initTime);
    
###############################################################################
# Initialise Station Arrays
###############################################################################

spaceTelescopes = [sc1]
groundTelescopes = [ALMA, LMT, SMA, APEX, JCMT, SPT, SMT, NOEMA, HAY, IRAM]
groundStations = [Haleakala, Lasilla, Achaea, Perth]

###############################################################################
# Simulation
###############################################################################

# Execute time loop (i.e. run simulation)
spaceTelescopes, groundTelescopes, groundStations, simTime = TimeLoop(initTime, \
                simLength, timeStep, spaceTelescopes, groundTelescopes, \
                groundStations, obsFreq, sourceRa, sourceDec, dutyCycle, \
                intTime, allsky) 
    
###############################################################################
# Plot Figures
###############################################################################

# Plot space telescope orbits
Figures.OrbitPlot(spaceTelescopes)

# Plot (u,v) coverage
Figures.UvPlot(spaceTelescopes, groundTelescopes, allsky, 1)

# Plot attitude sphere (user can control which elements (e.g. Earth, Sun, Moon,
# antenna, etc.) are included in plot with additional arguments)
Figures.AttitudeSphere(spaceTelescopes, 0, 120, 30)

# Plot incidence angle of Sun on solar panels
Figures.SolarPanelIncidence(spaceTelescopes, simTime, 0)

# Plot elevation of space telescope from ground station
Figures.GroundStationElevation(spaceTelescopes, groundStations, simTime, 0)

# Plot Earth map with space telescope ground track and ground station visibility
Figures.ElevationEarthMap(spaceTelescopes, groundStations, 0)

###############################################################################
# Optimisation
###############################################################################

# Run Optimisation function to determine optimal position for a star tracker
# with a Sun and Earth minimum exclusion angle of 30 degrees.
results = Optimisation(spaceTelescopes, 0, 30, 30, 0)
